
from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2024-09",
    fileinfo = "File to set workflows."
)

__all__ = """
    scope
    Jump
    jump
    Workflow
    Switch
    switch
""".split()

from .timing import Timer
from .exception import Error
from .decorators import alias

class Jump(object):
    """
    Creates a Jump Error to escape scopes. 

    Examples::
        >>> with scope("test"), jump:
        ...     # inside codes
        ... 
        # nothing, the inside codes do not run
        >>> with scope("test"), Jump(False) as stop:
        ...     print("Part I")
        ...     stop()
        ...     print("Part II")
        ... 
        Part I
    """
    def __init__(self, jump=None): self.jump = True if jump is None else jump
    def __enter__(self):
        def dojump(): raise Error("Jump")("Jump by class 'Jump'. ")
        if self.jump: dojump()
        else: return dojump
    def __exit__(self, *args): pass
    def __call__(self, condition): return Jump(condition)
    
def scope(name, log_on_screen=True):
    """
    An allias of timer to better organize the codes, use .exit() to exit the scope. 
    
    Args:
        name (str): the name of the scope, used to display. 
        log_on_screen (bool): whether to show the time span or not. 

    Examples::
        >>> with scope("test"):
        ...     # inside codes
        ... 
        # some outputs
        [scope test takes 0.001s]
        >>> with scope("this") as s:
        ...     print("Part I")
        ...     s.exit()
        ...     print("Part II")
        ...
        Part I
        >>> with scope("this again", False) as s:
        ...     print("Part I")
        ...     print("Part II")
        ...
        Part I
        Part II
        >>> print(s.recorded_time)
        2.86102294921875e-06
    """
    return Timer("scope " + str(name), timing=True, log_on_screen=log_on_screen)

jump = Jump()
"""
The jumper, one can use it along with `scope`(or `Timer`) to jump a chunk of codes. 
"""

class Workflow:
    """
    A structure to create a series of workflow. 
    
    Note:
        Remember to manually add a behaviour for each block: 
            '*.force_run' force the block to run, without checking the workflow.
            '*.force_skip'/'*.force_jump' force the block to be skipped, without checking the workflow.
            '*.use_tag'/'*.run_as_workflow' runs the block following the workflow schedule if one tag name is provided.
            '*.all_tags'/'*.run_if_all_tags_in_workflow' runs the block when all given tags are defined in the workflow. 
            '*.any_tag'/'*.run_if_any_tag_in_workflow' runs the block when at least one tag is defined in the workflow. 
        Fogetting to add the behaviour would result in an automatic run of blocks. See the example for details of bahaviours. 
    
    Args:
        *args: the list of scope names to run. 

    Examples::
        >>> run = Workflow("read data", "run method", "visualization")
        ... with run("read data"), run.use_tag:
        ...     print(1, end='')
        ... with run("pre-processing"), run.use_tag:
        ...     print(2, end='')
        ... with run("pre-processing", "run method"), run.all_tags:
        ...     print(3, end='')
        ... with run("visualization"), run.force_skip:
        ...     print(4, end='')
        ... 
        1[read data takes 0.000022s]
    """
    def force_jump(self): ...
    def run_as_workflow(self): ...
    def all_tags(self): ...
    def use_tag(self): ...
    def any_tag(self): ...
    def __init__(self, *args, verbose=True): self.workflow = args; self.verbose=verbose
    def __call__(self, *keys):
        if len(keys) == 1 and isinstance(keys[0], (list, tuple)): keys = keys[0]
        self.keys=keys
        return Timer(','.join(keys), timing=self.verbose)
    def __getattr__(self, *k): return self(*k)
    def __getitem__(self, *k): return self(*k)
    @property
    def force_run(self): return Jump(False)
    @alias("force_jump")
    @property
    def force_skip(self): return Jump(True)
    @alias("run_as_workflow", "all_tags", "use_tag")
    @property
    def run_if_all_tags_in_workflow(self):
        return Jump(any(k not in self.workflow for k in self.keys))
    @alias("any_tag")
    @property
    def run_if_any_tag_in_workflow(self):
        return Jump(all(k not in self.workflow for k in self.keys))

class Switch:
    """
    The replacement of C-styled switch in Python. 
    One can use the scope in the following 2 ways where #2 is more recommanded. 
    Example #1::
        >>> from pycamia import Switch
        >>> obj = 2
        >>> with Switch(obj) as case:
        ...     with case(1) as this:
        ... 	    assert this
        ... 	    print("hi")
        ...
        >>>     with case(2) as this:
        ... 	    assert this
        ... 	    print('bye')
        ...
        bye
    Example #2 (NOTE: switch[=Switch()] here is in lower cases. )::
        >>> from pycamia import switch
        >>> obj = 3
        >>> with switch:
	    ...     assert obj == 1
	    ...     print("hi")
        ...
        >>> with switch:
	    ...     assert obj == 2
	    ...     print('bye')
        ...
        >>> switch.close()
        Traceback (most recent call last):
            [...]
        pycamia.exception.SwitchError: No case matched. 
    """
    def __init__(self, variable=None, as_value=False):
        self.value = variable
        self.values = []
        self.as_value = as_value
        self.matched = False
    def __call__(self, value):
        self.values.append(value)
        if value == self.value:
            self.matched = True
        return Switch(value == self.value, as_value = True)
    def __enter__(self): return self.value if self.as_value else self
    def __exit__(self, error_type, error_msg, traceback):
        if error_type is None:
            if self.values: self.close()
            else: self.matched = True
            return True
        if error_type == Error("Jump") or error_type == AssertionError: return True
    @alias("end_switch")
    def close(self):
        if not self.matched:
            if self.value is not None:
                error_msg = f"Unrecognized case {self.value}"
                if self.values: error_msg += f", the available options are {self.values}. "
                else: error_msg += '. '
            else: error_msg = "No case matched. "
            raise Error("Switch")(error_msg)

switch = Switch()



from .manager import info_manager

__info__ = info_manager(
    project = 'PyCAMIA',
    package = '<main>',
    author = 'Yuncheng Zhou',
    create = '2021-12',
    version = '1.0.40',
    update = '2024-05-22 17:35:24',
    contact = 'bertiezhou@163.com',
    keywords = ['environment', 'path', 'touch'],
    description = 'The main package and a background support of project PyCAMIA. ',
    requires = ['tqdm', 'psutil']
).check()
__version__ = '1.0.40'

from .environment import get_environ_vars, get_environ_globals, get_environ_locals, update_locals_by_environ, get_args_expression, get_reference_line, get_declaration, EnvironVars #*
from .exception import touch, crashed, avouch, Error #*
from .functions import empty_function, const_function, identity_function #*
from .inout import no_out, no_print, SPrint, StrIO, input_with_timeout #*
from .manager import info_manager, Hyper, hypers, Version, Args, args #*
from .timing import time_this, Timer, periodic, periodic_run, periodic_call, run_later #*
from .workflow import scope, Jump, jump, Workflow, Switch, switch #*
from .decorators import alias, decorator, restore_type_wrapper #*
from .listop import prod, cumsum, cumprod, cartesian_prod, argmin, argmax, min_argmin, max_argmax, kth_biggest, kth_smallest, median, flatten_list, item, to_list, to_tuple, to_set, map_ele, sublist, arg_tuple, arg_extract, count, unique, infinite_itemize, cat_generator, param_join #*
from .loopop import Collector, iterate #*
from .strop import is_digits, is_alphas, is_snakename, get_digits, get_alphas, get_snakename, get_snakenames, str_len, str_slice, find_all, enclosed_object, tokenize, token_replace, columns #*
from .pythonop import sorted_dict_repr, dict_parse, set_indent_len, get_indent_len, tab_to_indent, no_indent, with_indent, get_num_indent, python_lines, python_line, execblock, add_lineno #*
from .system import path, Path, path_list, PathList, set_black_list, get_black_list, curdir, pardir, homedir, rootdir, pwd, ls, cp, copy, mv, move, merge, rename, is_valid_command, Shell, get_volumes, ByteSize #*
from .logging import logging, start_logging #*
from .math import GCD, isint, rational, factorial #*
from .structures import struct, odict #*
from .more import once, void, random_seed #*
from . import system as pth




from pycamia import info_manager

__info__ = info_manager(
    project = "PyZMyc",
    package = "pyoverload",
    fileinfo = "Useful tools for decorators."
)

__all__ = """
    raw_function
    decorator
    get_environ_vars
""".split()

import os, sys
from functools import wraps

try:
    import ctypes
except ModuleNotFoundError:
    ctypes = None
else:
    if hasattr(ctypes, "pythonapi") and \
       hasattr(ctypes.pythonapi, "PyFrame_LocalsToFast"): pass
    else: ctypes = None

def raw_function(func):
    if hasattr(func, "__func__"):
        return func.__func__
    return func

def _get_wrapped(f):
    while hasattr(f, '__wrapped__'): f = f.__wrapped__
    return f

def decorator(wrapper_func):
    if not callable(wrapper_func): raise TypeError(f"@decorator wrapping a non-wrapper: {wrapper_func}")
    def wrapper(*args, **kwargs):
        if len(args) == 1 and callable(args[0]):
            func = args[0]
            raw_func = raw_function(func)
            func_name = f"{raw_func.__name__}[{wrapper_func.__qualname__.split('.')[0]}]"
            wrapped_func = wraps(raw_func)(wrapper_func(raw_func))
            wrapped_func.__name__ = func_name
            wrapped_func.__doc__ = raw_func.__doc__
            # return wrapped_func
            if 'staticmethod' in str(type(func)): trans = staticmethod
            elif 'classmethod' in str(type(func)): trans = classmethod
            else: trans = lambda x: x
            return trans(wrapped_func)
        return decorator(wrapper_func(*args, **kwargs))
    return wraps(wrapper_func)(wrapper)

def _mid(x): return x[1] if len(x) > 1 else x[0]
def _rawname(s): return _mid(str(s).split("'"))

stack_error = lambda x, ext: TypeError(f"Unexpected function stack for {x}, please contact the developer for further information (Error Code: E001*). {ext}")

# def _get_frames():
#     frames = []
#     frame = sys._getframe()
#     fname = frame.f_back.f_code.co_name
#     while frame is not None:
#         frame_file = _rawname(frame)
#         if frame_file.startswith('<') and frame_file.endswith('>') and frame_file != '<stdin>':
#             frame = frame.f_back
#             continue
#         frames.append(frame)
#         if len(frames) >= 4: return frames[2:]
#         frame = frame.f_back
#     raise stack_error(fname)

# def get_environ_locals():
#     _, client_frame = _get_frames()
#     return client_frame.f_locals

# def get_environ_globals():
#     _, client_frame = _get_frames()
#     return client_frame.f_globals

def _get_frames(i = [2, 3], key=''):
    """
    Get frames in stack. 
    By default: it gets frame of the function calling get_environ (function frame) and the frame calling this function (client frame). 
    Returns: function frame, client frame
    """
    frames = []
    frame = sys._getframe()
    fname = frame.f_back
    if isinstance(i, int): i = [i]
    if i is not None:
        if len(i) == 0: raise IndexError("Invalid index for _get_frames")
        max_i = max(i)
    while frame is not None:
        frame_file = frame.f_code.co_filename
        if frame_file.startswith('<') and frame_file.endswith('>') and frame_file != '<stdin>':
            frame = frame.f_back
            continue
        if i is None:
            if frame.f_code.co_name == key: frames.append(frame)
        else:
            frames.append(frame)
            if len(frames) >= max_i + 1:
                domain = [frames[j] for j in i]
                if key != '': domain = [f for f in domain if f.f_code.co_name == key]
                return domain if len(domain) > 1 else domain[0]
        frame = frame.f_back
    if i is not None or len(frames) == 0:
        try: f_all = _get_frames(-1)
        except: raise stack_error(fname, f"\n_get_frames({i}) got stack: \n" + '\n'.join(map(str, frames)))
        raise stack_error(fname, "\nComplete stack: \n" + '\n'.join(map(str, f_all)) + f"\n_get_frames({i}) got stack: \n" + '\n'.join(map(str, frames)))
    return frames

class EnvironVars():
    
    def __init__(self, frame): self.frame = frame

    def get(self, name, default=None):
        res = self.frame.f_locals.get(name, self.frame.f_globals.get(name, default))
        if res is None: raise AttributeError(f"No variable {name} found in the environment. ")
        return res

    def set(self, name, value, in_dict=None):
        if not in_dict: in_dict = 'local'
        if in_dict.lower().startswith('loc'): self.frame.f_locals[name] = value
        else: self.frame.f_globals[name] = value
        if ctypes is not None:
            ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(self.frame), ctypes.c_int(0))
    
    def update(self, dic, in_dict=None):
        for k, v in dic.items():
            if not in_dict: in_dict = 'local'
            if in_dict.lower().startswith('loc'): self.frame.f_locals[k] = v
            else: self.frame.f_globals[k] = v
        if ctypes is not None:
            ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(self.frame), ctypes.c_int(0))
        
    def __contains__(self, key): return key in self.frame.f_locals or key in self.frame.f_globals
    def __getitem__(self, key): return self.get(key)
    def __setitem__(self, key, value): return self.set(key, value)
    def __getattr__(self, key):
        if key in self.__dict__: return super().__getattr__(key)
        return self.get(key)
    def __setattr__(self, key, value):
        if key == 'frame': return super().__setattr__(key, value)
        return self.set(key, value)
    
    @property
    def locals(self): return self.frame.f_locals
    
    @property
    def globals(self): return self.frame.f_globals
    
    @property
    def all(self):
        all = self.frame.f_globals.copy()
        all.update(self.frame.f_locals)
        return all
    
def get_environ_vars(offset=0, pivot=''):
    client_frame = _get_frames(3) # offset of frame
    if pivot: client_frame = _get_frames(None, key=pivot)
    if isinstance(client_frame, list): client_frame = client_frame[-1]
    for _ in range(offset):
        client_frame = client_frame.f_back
    return EnvironVars(client_frame)

def update_locals_by_environ():
    module_frame, client_frame = _get_frames()
    vars_set = client_frame.f_locals.copy()
    vars_set.update(module_frame.f_locals)
    module_frame.f_locals.update(vars_set)

class StrIO:
    def __init__(self, file_name = os.path.abspath('.null')):
        self._str_ = None
        self._file_ = open(file_name, 'w+')
        self.file_name = file_name
        self.fileno = self._file_.fileno
    def write(self, s): self._file_.write(s)
    def __str__(self):
        if self._str_ is not None: return self._str_
        self._file_.seek(0)
        self._str_ = self._file_.read()
        self.close()
        return self._str_
    def split(self, c=None): return str(self).split(c)
    def string(self): return str(self)
    def close(self):
        self._file_.close()
        if self.file_name == os.path.abspath('.null'):
            os.remove(self.file_name)
            self._file_ = None
    
def get_args_expression():
    module_frame, client_frame = _get_frames()
    func_name = module_frame.f_code.co_name
    if os.path.exists(client_frame.f_code.co_filename):
        with open(client_frame.f_code.co_filename) as fp:
            for _ in range(client_frame.f_lineno-1): fp.readline()
            l = fp.readline()
            if func_name not in l:
                raise TypeError(f"Cannot find function name `{func_name}` in {client_frame.f_code.co_filename} line {client_frame.f_lineno}:\n\t{l}. \n" +
                    "Problem occurs in code stack, please contact the developer for further information (Error Code: E002*). ")
            exp = l.split(func_name)[-1].split(';')[0].strip()
            if exp.startswith('('): exp = exp[1:]
            if exp.endswith(')'): exp = exp[:-1]
            return exp
    else: return "<unreachable arg expression>"

def get_declaration(func):
    func_code = func.__code__
    func_name = func.__name__
    if os.path.exists(func_code.co_filename):
        with open(func_code.co_filename) as fp:
            for _ in range(func_code.co_firstlineno - 1): fp.readline()
            i = func_code.co_firstlineno
            while True:
                l = fp.readline()
                if func_name not in l:
                    if not l.strip() or l.strip().startswith('@') or l.strip().startswith('#'): i += 1; continue
                    raise TypeError(f"Cannot find function name `{func_name}` in {func_code.co_filename} line {i}:\n\t{l}. \n" +
                        "Problem occurs in code stack, please contact the developer for further information (Error Code: E003*). ")
                dec_line = l
                break
    else:
        ss = StrIO()
        oldout = sys.stdout
        sys.stdout = ss
        help(func)
        sys.stdout = oldout
        dec_line = [l for l in ss.split('\n') if len(l) > 0 and 'Help' not in l][0]
    if not dec_line: return "<unreachable declaration>"
    dec = dec_line.strip().split(' ', 1)[-1].rstrip(':')
    return dec




import ctypes
import numpy as np
import micomputing as mc
from pycamia import Path

dll = ctypes.cdll.LoadLibrary(Path(__file__).parent/'micfunctions.so')
distance_map_func = dll.distance_map
distance_map_func.argtypes = [np.ctypeslib.ndpointer(dtype=np.int32,ndim=1,flags="C_CONTIGUOUS"),
                              np.ctypeslib.ndpointer(dtype=np.float32,ndim=1,flags="C_CONTIGUOUS"),
                              ctypes.c_int,
                              np.ctypeslib.ndpointer(dtype=np.int32,ndim=1,flags="C_CONTIGUOUS"),
                              np.ctypeslib.ndpointer(dtype=np.float32,ndim=1,flags="C_CONTIGUOUS")]

masks = ((np.stack(np.meshgrid(np.arange(100), np.arange(100), indexing='ij'), 0) - np.array((20, 20)).reshape((2, 1, 1))) ** 2).sum(0) < 100
mask_sent = masks.flatten().astype(np.int32)
dismap_get = np.zeros_like(mask_sent).flatten().astype(np.float32)
size_in = np.array((1, 100, 100)).astype(np.int32)
spacing_in = np.array((1, 1)).astype(np.float32)
distance_map_func(mask_sent, dismap_get, 3, size_in, spacing_in)
dismap_get = dismap_get.reshape(masks.shape)

mc.plot.subplots(2)
mc.plot.imshow(masks)
mc.plot.imshow(dismap_get)
mc.plot.show()




try: from pycamia import info_manager
except ImportError: print("Warning: pycamia not loaded. ")

__info__ = info_manager(
    project = "PyCAMIA",
    package = "pyoverload",
    author = "Yuncheng Zhou", 
    create = "2023-09",
    fileinfo = "overload functions",
    requires = ""
)

__all__ = """
    overload
    OverloadError
    
    deprecated override
""".split()

deprecated = ...
def override(*_a, **_k): raise NotImplementedError("""
[WARNING] pyoverload has upgraded into pyoverload II, please read the new README.md for detailed information.
To simply get your previous codes running, do this: 
from pyoverload.old_version_files_deprecated.override import overload
from pyoverload.old_version_files_deprecated.typehint import *
If you are using the 'decorator fashion' (@overload for all imp's) of pyoverload I, use the following imports of new functions will work.
from pyoverload import overload
from pyoverload import typehint as params
from pyoverload.old_version_files_deprecated.typehint import * # Special types such as Int, etc. still need this line of import. 
Do change the capitalized types into lower_cased ones, and @params into @typehint later.
""")

from functools import wraps
from .typehint import typehint, decorator
from .typehint import TypeHintError, HintTypeError, get_virtual_declaration
from .typings import *

with __info__:
    from pycamia import Path

class OverloadError(Exception): ...

def wrapped(f):
    while hasattr(f, '__wrapped__'):
        f = f.__wrapped__
    return f

class Overloader:

    def __init__(self):
        """@overload decorator for overloading a function with multiple usages.
        The usages are given by annotations after colons.
        
        @overload find the first suitable function to run according to the order of definition.

        Examples::
            >>> @overload
            ... def func(self, pos1, pos2: str="def_val1", / , arg3: (int, float)=3, arg4=[], *args: int, kwarg5: str='extra', **kwargs) -> str:
            ...     ... # functional content
            ...
            >>> @overload
            ... def func(self, pos1, pos2: int=1, / , arg3: (int, float)=3, arg4=[], *args: int, kwarg5: str='extra', **kwargs) -> str:
            ...     ... # functional content
            ...
            >>> fun(self, 10, 20) # call the second function
        """
        self.prev_var_name = None
        self.funcset_collected = {}

    @decorator
    def __register__(self, func):
        wrapped_func = wrapped(func)
        func_loc = Path(wrapped_func.__code__.co_filename)
        var_name = '.'.join([func_loc.parent.name, func_loc.name, (wrapped_func.__func__ if isinstance(wrapped_func, method) else wrapped_func).__name__])
        if var_name.endswith('._'):
            if self.prev_var_name is None:
                raise NameError("Cannot use '_' as a function name of the first @overload.")
            var_name = self.prev_var_name
            var_name_candidates = [var_name]
        else:
            var_name_candidates = [var_name]
            parts__ = var_name.split('__')
            if len(parts__) >= 2 and parts__[-1]: var_name_candidates.append('__'.join(parts__[:-1]))
            if len(parts__) >= 3 and parts__[-2] and not parts__[-1]: var_name_candidates.extend(['__'.join(parts__[:-2] + ['']), '__'.join(parts__[:-2])])
            parts_ = var_name.split('_')
            if len(parts_) >= 3 and parts_[-2] and not parts_[-1]: var_name_candidates.extend(['_'.join(parts_[:-2] + ['']), '_'.join(parts_[:-2])])
        
        if not ':typehint' in func.__name__: func = typehint(func, check_return=True)
        func_index = None
        for v in var_name_candidates:
            if v in self.funcset_collected:
                if v != var_name: func_index = len(self.funcset_collected[v])
                self.funcset_collected[v].append(func)
                var_name = v
                break
        else:
            if var_name.endswith('__default__'): var_name = var_name[:-len('__default__')]
            elif var_name.endswith('__0__'): var_name = var_name[:-len('__0__')]
            self.funcset_collected[var_name] = [func]
        self.prev_var_name = var_name
        @wraps(func)
        def wrapper(*args, **kwargs):
            return self.__call__(*args, __cur__func__ = (var_name, func_index), **kwargs)
        return wrapper

    def __call__(self, *args, __cur__func__=None, **kwargs):
        if len(args) == 1 and not kwargs and not __cur__func__ and isinstance(args[0], callable):
            outer_func = self.__register__(args[0])
            var_name = self.prev_var_name
            declarations = '\n    '.join([get_virtual_declaration(wrapped(f)) for f in self.funcset_collected[var_name]])
            func_doc = [f.__doc__ for f in self.funcset_collected[var_name] if f.__doc__ is not None]
            if func_doc: func_doc = func_doc[0]
            else: func_doc = ''
            outer_func.__doc__ = f"""
@overload registered function at: '{var_name}', with the following usages:
    {declarations}
{func_doc}
            """
            return outer_func
        if __cur__func__ is None: raise TypeError("Invalid @overload sub-function without keyword argument '__cur__func__'")
        if __cur__func__[1] is not None:
            # direct call
            func = self.funcset_collected[__cur__func__[0]][__cur__func__[1]]
            try: return func(*args, **kwargs)
            except (HintTypeError, TypeHintError) as e:
                args_str = ', '.join([repr(x) for x in args] + ['='.join((str(x[0]), repr(x[1]))) for x in kwargs.items()])
                raise OverloadError(f"Failed in direct call of {wrapped(func).__name__}() with arguments {args_str}. The usage is:\n" + get_virtual_declaration(wrapped(func)) + f": [{e.__class__.__name__}] {str(e)}")
        func_list = self.funcset_collected[__cur__func__[0]]
        
        failed_func_list = []
        for i, func in enumerate(func_list):
            # return func(*args, **kwargs)
            try: return func(*args, **kwargs)
            except (HintTypeError, TypeHintError) as e:
                failed_func_list.append((func, f": [{e.__class__.__name__}] {str(e)}"))
        args_str = ', '.join([repr(x) for x in args] + ['='.join((str(x[0]), repr(x[1]))) for x in kwargs.items()])
        raise OverloadError(f"No {__cur__func__[0]}() matches arguments {args_str}. All available usages are:\n" + '\n'.join([get_virtual_declaration(wrapped(f)) + e for f, e in failed_func_list]))

overload = Overloader()




from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "batorch",
    fileinfo = "The inherited tensor from 'torch' with batch.",
    requires = ["pycamia", "torch", "pynvml"]
)

__all__ = """
    CPU
    AutoDevice
    SleepingDevice
    free_memory_amount
    all_memory_amount
""".split()

import os, torch
with __info__:
    import pynvml, psutil
    from pycamia import alias, ByteSize

GPU_priority = ['cuda']#, 'mps']
is_available = dict(
    cuda = hasattr(torch, 'cuda') and torch.cuda.is_available(),
    mps = hasattr(torch.backends, 'mps') and torch.backends.mps.is_available(),
)
GB = 1.074e+9
warning_free_memory_threshold = eval(os.environ.get('CUDA_RUN_MEMORY', '5')) # Run with at lease *GB

free_memory_amount = None
all_memory_amount = None

@alias("free_memory_amount", amount="free")
@alias("all_memory_amount", amount="total")
def memory_amount(device_id, amount="total"):
    if not isinstance(device_id, str):
        if device_id == -1: device_id = 'cpu'
        else: device_id = f'cuda:{device_id}'
    if device_id == 'cpu':
        info = psutil.virtual_memory()
    elif device_id == 'mps':
        return dict(free=0, total=0)[amount]
    elif device_id.startswith('cuda'):
        pynvml.nvmlInit()
        h = pynvml.nvmlDeviceGetHandleByIndex(int(device_id.split(':')[-1]))
        info = pynvml.nvmlDeviceGetMemoryInfo(h)
    return getattr(info, amount)

def device_by_id(device_id):
    return torch.device(device_id)

class AutoDevice:
    """
    Auto find and set a device. 
    
    Args:
        fixed_device (torch.device): use a fixed device. 
        auto (bool): decide whether the device is auto selected. 
        required_memory (int): an estimated memory use which helps the selection, in GB. 
        verbose (bool): whether to print warnings or not.

    Examples::
        >>> gpu = AutoDevice()
        >>> tensor.to(gpu.device)
        <tensor>
        >>> gpu.auto = False
        >>> gpu(tensor)
        <tensor>
    """
    def __init__(self, *devices, required_devices=1, auto=True, required_memory=None, verbose=True, always_proceed=False):
        if len(devices) == 0: devices = ['cpu']
        self.required_devices = required_devices
        self.required_memory = required_memory
        self.always_proceed = always_proceed
        self.verbose = verbose
        self.auto = auto
        self.available_gpu_ids = []
        self._working_devices = None
        self.default_devices = devices

    @property
    def working_devices(self):
        if self.auto: self.auto_select()
        elif self._working_devices is None:
            self._working_devices = [f"cuda:{d}" if isinstance(d, int) else str(d) for d in self.default_devices]
        return self._working_devices

    @alias("device_id")
    @property
    def main_device_id(self): return self.working_devices[0]
        
    @alias("device")
    @property
    def main_device(self): return device_by_id(self.main_device_id)

    @property
    def working_memory(self):
        return ByteSize(free_memory_amount(self.main_device_id))

    def auto_select(self):
        self.auto = False
        for gpu_family in GPU_priority:
            if gpu_family == 'cuda' and is_available['cuda']:
                self.available_gpu_ids = list(range(torch.cuda.device_count()))
                available_gpus_memory = [free_memory_amount(i) for i in self.available_gpu_ids]

                threshold = warning_free_memory_threshold if self.required_memory is None else self.required_memory / self.required_devices / 2
                sorted_memory = sorted(enumerate(available_gpus_memory), key=lambda x: -x[1])
                candidates = [(i, m) for i, m in sorted_memory if m > threshold * GB][:self.required_devices]
                if len(candidates) > 0: most_available_gpus, most_available_gpu_memory = zip(*candidates)
                else: 
                    print(f"Warning: no gpu device is available (with memories {available_gpus_memory}), reset environment variable CUDA_RUN_MEMORY to change the memory required...")
                    print(f"Most memory: {sorted_memory[0][1] / GB:.5} GB in GPU {sorted_memory[0][0]}. ")
                    if not self.always_proceed:
                        tag = input("Do you want to proceed with CPU? [yes/no/y/n]:")
                        if 'y' not in tag.lower(): raise RuntimeError("Not enough GPU resource.")
                    continue
                if self.required_memory and sum(most_available_gpu_memory) < self.required_memory and self.verbose:
                    print(f"Warning: all remaining memory of gpu devices is not enough (on: {most_available_gpus})...")
                    print(f"Total free memory: {sum(most_available_gpu_memory) / GB:.5} GB. ")
                    if not self.always_proceed:
                        tag = input("Do you want to proceed? [yes/no/y/n]:")
                        if 'y' not in tag.lower(): raise RuntimeError("Not enough free memory left.")
                elif len(most_available_gpus) < self.required_devices and self.verbose:
                    print(f"Warning: not enough gpus available (available: {most_available_gpus})...")
                    print(f"Total free memory: {sum(most_available_gpu_memory) / GB:.5} GB. ")
                    if not self.always_proceed:
                        tag = input("Do you want to proceed? [yes/no/y/n]:")
                        if 'y' not in tag.lower(): raise RuntimeError("Not enough free devices left.")
                if self.verbose: print(f"Setting working devices: {most_available_gpus}, main device: {most_available_gpus[0]}. proceeding...")
                self._working_devices = [f"cuda:{i}" for i in most_available_gpus]
                break
            elif gpu_family == 'mps' and is_available['mps']:
                self._working_devices = ['mps']
                self.required_devices = 1
                break
        else:
            if self.verbose: print("Warning: cannot find any gpu, using cpu instead. ")
            self._working_devices = ['cpu']
            self.required_devices = 1
    
    def turn_on(self):
        if not self.auto:
            self.auto = True
            self.auto_select()
        return self
    
    def turn_off(self):
        if self.auto: self.auto = False
        return self
    
    def __eq__(self, other):
        if isinstance(other, torch.device): return self.main_device == other
        if isinstance(other, AutoDevice): return self.main_device == other.main_device
        return self == other
    
    def __call__(self, x):
        mdevice = self.main_device
        if isinstance(x, torch.Tensor):
            if x.device == mdevice: return x
            return x.to(mdevice)
        elif isinstance(x, torch.nn.Module):
            if len(self.working_devices) <= 1: return x.to(mdevice)
            return torch.nn.DataParallel(x.to(mdevice), device_ids=[int(x.split(':')[-1]) for x in self.working_devices])
        elif isinstance(x, torch.optim.Optimizer):
            if len(self.working_devices) <= 1: return x
            return torch.nn.DataParallel(x, device_ids=[int(x.split(':')[-1]) for x in self.working_devices])
        else: return torch.tensor(x, device=mdevice)
        return x
    
class SleepingDevice(AutoDevice):
    def __init__(self, *devices, required_devices=1, auto=False, required_memory=None, verbose=False):
        super().__init__(*devices, required_devices=required_devices, auto=auto, required_memory=required_memory, verbose=verbose) 

CPU = SleepingDevice(torch.device("cpu"))
if is_available['mps']:
    __all__.append('MPS')
    MPS = SleepingDevice(torch.device("mps"))
if is_available['cuda']:
    __all__.extend(['GPU', 'GPUs'])
    GPU = SleepingDevice(torch.device("cuda:0"))
    GPUs = [SleepingDevice(torch.device(f"cuda:{i}")) for i in range(torch.cuda.device_count())]




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2023-08",
    fileinfo = "File of loop operations. "
)

__all__ = """
    Collector
    iterate
""".split()

import time, builtins
from .inout import StrIO
from .strop import tokenize
from .listop import median
from .exception import avouch, touch
from .decorators import alias
from datetime import datetime, timedelta
from threading import Thread
from typing import Iterable

with __info__:
    use_nput = False
    try:
        from pynput import keyboard as kb
        use_nput = True
    except ImportError: ...

class Collector:
    def __init__(self, max_storage=None):
        self._collection = {}
        self._format = {}
        self._max_storage = max_storage
    def register(self, name, value, format='.4f/-.4f'):
        """format: mean format / std format. '-' for omission of '0'."""
        if len(name.split(':')) == 2:
            name, format = name.split(':')
        name = name.strip()
        format = format.strip()
        if format != '':
            if '/' not in format: format = format + '/' + format
            avouch(len(format.split('/')) == 2, f"Invalid format: {format}, should be like '.4f' or '2d/.4f'.")
            format = format.split('/')
        if name not in self._collection:
            self._collection[name] = []
            avouch(name not in self._format, "Collector having format for values never collected, Please contact the developer for more information. (Error Code: C388)")
            self._format[name] = format
        elif name not in self._format or self._format[name] == '':
            self._format[name] = format
        else:
            avouch(self._format[name] == format or format == '', "Cannot assign different format for the same collection in class 'Collector'. ")
        self._collection[name].append(value)
        if self._max_storage is not None and len(self._collection[name]) > self._max_storage: self._collection[name].pop(0)
    def __contains__(self, name):
        return name in self._collection
    def __getitem__(self, name):
        if name in self._collection: return self._collection[name]
        raise NameError(f"No element registered as {name}. ")
    def wilcoxon(self, other, name):
        from scipy.stats import wilcoxon
        if isinstance(other, Collector): other = other.as_numpy(name)
        return wilcoxon(self.as_numpy(name), other)[1]
    @alias("as_array")
    def as_numpy(self, name):
        import numpy as np
        return np.array(self[name])
    def as_str_tuple(self, name, ignore_if_not_collected=False):
        if ignore_if_not_collected and name not in self._collection: return ('', '')
        a = self.as_numpy(name)
        format = self._format.get(name, '')
        if format == '': format = ['', '']
        ret = []
        for f, v in zip(format, (a.mean(), a.std())):
            if f.startswith('-'):
                ret.append(f"{{var:{f[1:]}}}".format(var=v).lstrip('0'))
            else:
                ret.append(f"{{var:{f}}}".format(var=v))
        return tuple(ret)
    def as_pm(self, name, ignore_if_not_collected=False):
        if ignore_if_not_collected and name not in self._collection: return ''
        m, s = self.as_str_tuple(name)
        return m + '±' + s
    def as_latex(self, *names, ignore_if_not_collected=False):
        output_cells = []
        if len(names) == 0: names = self._collection.keys()
        for name in names:
            if name not in self._collection:
                if ignore_if_not_collected: continue
                output_cells.append('-'); continue
            m, s = self.as_str_tuple(name)
            output_cells.append(m + '$\pm$' + s)
        return ' & '.join(output_cells)

class HotKeyListener:
    keyboard = "asdfhgzxcv bqweryt123465=97-80]ou[ip<enter>lj'k;\\,/nm.<tab> `<backspace> <esc><cmd-r><cmd><shift><caps-lock><alt><ctrl><shift-r> <ctrl-r>  . * + \x1B   /\x03 -   01234567 89   <f5><f6><f7><f3><f8><f9> <f11> <f13> <f14> <f10>\x10<f12> <f15>\x05<home><page-up><delete><f4><end><f2><page-down><f1><left><right><down><up>"
    enumKeys = {k: v for k, v in enumerate(tokenize(keyboard, sep=[''], by='<>')[1:])}
    def __init__(self, *hotkeys, callback=lambda x:None):
        avouch(use_nput, "HotKeyListener cannot be used without package 'pynput'.")
        self.hotkeys = []
        self.str_reprs = []
        for ht in hotkeys: self.add_listen(ht)
        self.key_record = set()
        self.started = False
        self.detected = None
        self.callback = callback
        self.listener = None
    def on_keyboard(self, key, act):
        if isinstance(key, kb.KeyCode): key = key.char
        elif isinstance(key, kb._darwin.KeyCode): key = self.enumKeys[key.vk]
        elif isinstance(key, kb.Key):
            key_index = str(key.value).strip('<>')
            key_value_char = touch(lambda: self.enumKeys[int(key_index)], key_index)
            key = getattr(kb.Key, key.name, key_value_char)
        else: raise TypeError(f"Unrecognized type {type(key)} for a key {key}. ")
        if act == 'release':
            if key in self.key_record: self.key_record.remove(key)
        elif act == 'press':
            self.key_record.add(key)
            if self.key_record in self.hotkeys:
                self.detected = self.str_reprs[self.hotkeys.index(self.key_record)]
                self.callback(self)
    def add_listen(self, hotkey):
        ht = set()
        for k in hotkey.split('+'):
            k = k.lower()
            if len(k) > 1: ht.add(getattr(kb.Key, k, None))
            else: ht.add(k)
        self.hotkeys.append(ht)
        self.str_reprs.append(hotkey)
    def is_listening(self, hotkey):
        ht = set()
        for k in hotkey.split('+'):
            k = k.lower()
            if len(k) > 1: ht.add(getattr(kb.Key, k, None))
            else: ht.add(k)
        return ht in self.hotkeys
    def start(self):
        self.started = True
        if self.listener is None:
            self.listener = kb.Listener(
                on_press = lambda x: self.on_keyboard(x, 'press'),
                on_release = lambda x: self.on_keyboard(x, 'release'))
        self.listener.start()
    def stop(self):
        self.started = False
        self.listener.stop()
        self.listener = None
        self.detected = None

if not use_nput:
    latest_line = ''
    terminate = True
    def on_input():
        while True:
            l = input()
            if terminate: break
            if l == '': continue
            global latest_line
            latest_line = l

class InputKeywordListener:
    def __init__(self, *keywords, callback=lambda x:None):
        self.keywords = [x.lower() for x in keywords]
        self.detected = None
        self.callback = callback
        self.listener = None
        self.terminate = False
    def on_keyword(self):
        while True:
            global latest_line
            if latest_line.strip().lower() in self.keywords:
                self.detected = latest_line.strip().lower()
                self.callback(self)
                latest_line = ''
            if self.terminate: break
            time.sleep(1)
        self.started = False
        self.listener = None
    def add_listen(self, keyword): self.keywords.append(keyword)
    def is_listening(self, keyword): return keyword in self.keywords
    def start(self):
        global terminate
        if terminate:
            terminate = False
            Thread(target=on_input).start()
        self.started = True
        if self.listener is None:
            self.listener = Thread(target=self.on_keyword)
        self.listener.start()
    def stop(self):
        self.terminate = True
        global terminate
        terminate = True
        self.detected = None

def iterate(list_:(list, int), inner_loop=None, breakable=True, break_key=None):
    """
    iterate the input list_ with a visualized progress bar and a key break function 
        that can terminate the loop whenever a break key is entered.
        If we have package 'pynput' installed, we listen to the keyboard for hot keys. The break key is 'cmd+b' by default. 
        (If Darwin(mac) systems notify with security problem, please add your IDLE in the accessibility (辅助功能) list in system settings.)
        If we donot have access to the package, one have to enter the break_key string and press enter to send it, the break key is 'b' by default. 
        This might cause problem in formatting when a new line of output is printed during entering the break key, 
        hence it is recommended to use short break_keys and reduce the number of iterations for the loop using 'iterate'.
        P.S. One may find the program hard to exit in the end due to the 'input' function, feel free to press enter or close it the hard way. 

    Note: The input should be either an integer for a range or a list or tuple with fixed length. Generators should
        be cast into lists first. This ristriction is applied to avoid endless loading here when there are to many
        elements to be generated.
        
    Note2: Make sure to use `print` at the end of each iteration for correct progress bar. 
        Feel free to use nested loops with an inner loop without output.
        
    Args:
        list_ (list, int): iterative object to iterate by a 'for' loop. Using 'range' object for integer.
        inner_loop (int, iterable, callable):
            (int): the number of iteration in inner loop;
            (iterable): an iterable to loop through; or
            (callable): a function to create previous alternatives. 
                the function should accept arguments i and x (as in `enumerate(list_)`; x = i if `list_` is an integer):
                    i is the index of outer loop and x is the item.
                the function should return either the number of iterations or the data list to loop.
            Defaults to None.
            Using inner_loop will result in a two-element tuple for each iteration.
        breakable (bool): whether the user can interrupt the loop by 'break_key' or not.
        break_key (NoneType, str): the break key. e.g. 'ctrl+b'.

    Examples::
        >>> for i in iterate(20):
        ...     for _ in range(10000): ...
        ...     print(f'iteration {i}')
        ...
        00%     R--:--:--( -  s/it) |[15:46:48] iteration 0
            [...omitted...]
        99%████▉R 0:00:01(0.32s/it) |[15:46:52] iteration 19
        >>> for i, j in iterate(20, 10000):
        ...     print(f'epoch {i}, iteration {j}')
        ...
    """
    if breakable:
        if use_nput:
            if break_key is None: break_key = 'ctrl+b'
            listener = HotKeyListener(break_key)
        else:
            if break_key is None: break_key = 'b'
            listener = InputKeywordListener(break_key)
            print("Warning: 'iterate' cannot listen break keys without package 'pynput', ")
            print("builtin function 'input' will be used which demonds the user press enter after the default keyword 'b'. ")
            print("Note that if one is using darwin(mac) systems, 'pynput' may not be trusted, ")
            print("please add the IDLE in the accessibility (辅助功能) list of the system settings first. ")
    progress_chars = " ▏▎▍▌▋▊▉█"
    n_block_share = len(progress_chars) - 1
    progress_len = 10
    n_timespan_store = 20
    use_progress_bar = True
    if isinstance(list_, int):
        n_out = list_
        list_ = range(list_)
    else:
        if not hasattr(list_, '__len__'):
            use_progress_bar = False
            print(f"Warning: Function 'iterate' cannot predict progress with non-list iterative object {list_}, consider casting the argument first.")
        n_out = len(list_)
    if breakable: listener.start()
    iter_timespans = []
    inner_loop_timespans = []
    print(f"Loop starting at {datetime.now().strftime('%m/%d %H:%M:%S')}...")
    for i, x in enumerate(list_):
        iter_begin = datetime.now()
        if inner_loop is not None:
            if not isinstance(inner_loop, (int, Iterable)) and callable(inner_loop):
                inner_loop = inner_loop(i, x)
            use_inner_progress_bar = True
            if isinstance(inner_loop, int):
                n_in = inner_loop
                inner_loop = range(inner_loop)
            else:
                if not hasattr(inner_loop, '__len__'):
                    use_inner_progress_bar = False
                    print(f"Warning: Function 'iterate' cannot predict progress with non-list iterative object {inner_loop}, consider casting the argument first.")
                n_in = len(inner_loop)
        if inner_loop is None:
            if use_progress_bar:
                progress_pos = int(i * progress_len * n_block_share / n_out)
                progress_bar = '%02d%%'%(i * 100 // n_out)
                progress_bar += progress_chars[-1] * (progress_pos // n_block_share)
                progress_bar += progress_chars[progress_pos % n_block_share]
                progress_bar += ' ' * (progress_len - progress_pos // n_block_share - 1)
                if i > 0:
                    t_iter = median(iter_timespans)
                    remaining_time = int(t_iter * (n_out - i))
                    secs = remaining_time % 60
                    mins = (remaining_time // 60) % 60
                    hours = remaining_time // 3600
                    print_time = (iter_begin + timedelta(seconds=t_iter)).strftime("%H:%M:%S")
                    progress_bar += f"R{hours:2d}:{mins:02d}:{secs:02d}({t_iter:.2f}s/it) |[{print_time}]"
                else:
                    print_time = (iter_begin + timedelta(seconds=2)).strftime("%H:%M:%S")
                    progress_bar += f"R--:--:--( -  s/it) |[{print_time}]"
                print(progress_bar, end=" ")
                yield x
                iter_timespans.append((datetime.now() - iter_begin).total_seconds())
                if len(iter_timespans) > n_timespan_store: iter_timespans.pop(0)
                if breakable and listener.detected: break
        else:
            iter_timespans = []
            for j, y in enumerate(inner_loop):
                if use_progress_bar or use_inner_progress_bar:
                    progress_pos_out = int(i * progress_len / n_out)
                    progress_pos_in = int(j * progress_len * n_block_share / n_in)
                    progress_bar = '%02d%%'%(((i * n_in + j) * 100) / n_in / n_out)
                    if progress_pos_out * n_block_share < progress_pos_in:
                        progress_bar += progress_chars[-1] * (progress_pos_out - 1)
                        progress_bar += progress_chars[-2] * (progress_pos_out > 0)
                        progress_bar += progress_chars[-1] * (progress_pos_in // n_block_share - progress_pos_out)
                        progress_bar += progress_chars[progress_pos_in % n_block_share]
                        progress_bar += progress_chars[0] * (progress_len - progress_pos_in // n_block_share - 1)
                    else:
                        progress_bar += progress_chars[-1] * (progress_pos_in // n_block_share)
                        if progress_pos_out * n_block_share == progress_pos_in:
                            progress_bar += progress_chars[1]
                        else:
                            progress_bar += progress_chars[progress_pos_in % n_block_share]
                            progress_bar += progress_chars[0] * (progress_pos_out - progress_pos_in // n_block_share - 1)
                            progress_bar += progress_chars[1]
                        progress_bar += progress_chars[0] * (progress_len - progress_pos_out - 1)
                    if i > 0 or j > 0:
                        if i > 0: t_outer = median(inner_loop_timespans)
                        if j > 0: t_iter = median(iter_timespans)
                        if i == 0: t_outer = t_iter * n_in
                        if j == 0: t_iter = t_outer / n_in
                        remaining_time = int(t_iter * (n_in - j) + t_outer * (n_out - i - 1))
                        secs = remaining_time % 60
                        mins = (remaining_time // 60) % 60
                        hours = remaining_time // 3600
                        print_time = (iter_begin + timedelta(seconds=t_iter)).strftime("%H:%M:%S")
                        progress_bar += f"R{hours:2d}:{mins:02d}:{secs:02d}({t_iter:.2f}s/it) |[{print_time}]"
                    else:
                        print_time = (iter_begin + timedelta(seconds=2)).strftime("%H:%M:%S")
                        progress_bar += f"R--:--:--( -  s/it) |[{print_time}]"
                    print(progress_bar, end=" ")
                    yield x, y
                    iter_timespans.append((datetime.now() - iter_begin).total_seconds())
                    if len(iter_timespans) > n_timespan_store: iter_timespans.pop(0)
                    if breakable and listener.detected: break
            else:
                inner_loop_timespans.append(n_in * median(iter_timespans))
                if len(inner_loop_timespans) > n_timespan_store: inner_loop_timespans.pop(0)
                continue
            break
    else:
        if breakable: listener.stop()
        return
    print("-- manual termination of the loop --")
    if breakable: listener.stop()
    
builtin_print = builtins.print

class TaskProgress:
    
    def __init__(self, progress_len=10, n_timespan_store=20, show_progress_bar=True, breakable=None, break_key=None):
        self.progress_chars = " ▏▎▍▌▋▊▉█"
        self.n_block_share = len(self.progress_chars) - 1
        self.progress_len = progress_len
        self.n_timespan_store = n_timespan_store
        self.show_progress_bar = show_progress_bar
        self.breakable = use_nput if breakable is None else breakable
        self.sizes = []
        self.progresses = []
        self.show_progresses = []
        self.iter_estimations = []

        if break_key is None: break_key = 'ctrl+b' if use_nput else 'b'
        self.break_key = break_key
        self.break_keys = []
        self.should_break = []
        if breakable:
            self.listener = (HotKeyListener if use_nput else InputKeywordListener)(break_key)
            if not use_nput:
                builtin_print("Warning: 'TaskProgress' cannot listen break keys without package 'pynput', ")
                builtin_print("builtin function 'input' will be used which demonds the user press enter after the default keyword 'b'. ")
                builtin_print("Note that if one is using darwin(mac) systems, 'pynput' may not be trusted, ")
                builtin_print("please add the IDLE in the accessibility (辅助功能) list of the system settings first. ")

    def builtin_print(self, *values, sep=' ', end='\n', **kwargs):
        self.parent_print(self.prefix, *values, sep=sep, end=end, **kwargs)

    @property
    def prefix(self):
        total_progress = 0
        unit = 1
        for i, n in zip(self.progresses, self.sizes):
            if n < 0: break
            total_progress += i * 100 / n / unit
            unit *= n
        progress_bar = '%02d%%'%total_progress
        if total_progress >= 100: progress_bar = "1oo"
        if sum(self.show_progresses) >= 2 and sum(n > 1 for s, n in zip(self.show_progresses, self.sizes) if s) >= 2:
            progress_pos_out = int(total_progress * self.progress_len / 100)
        else: progress_pos_out = 0
        if sum(self.show_progresses) >= 1:
            last_act_loop = [i for i, x in enumerate(self.show_progresses) if x][-1]
            progress_pos_in = int(self.progresses[last_act_loop] * self.progress_len * self.n_block_share / self.sizes[last_act_loop])
        else: return "[not in loop]"
        if progress_pos_out * self.n_block_share < progress_pos_in:
            progress_bar += self.progress_chars[-1] * (progress_pos_out - 1)
            progress_bar += self.progress_chars[-2] * (progress_pos_out > 0)
            progress_bar += self.progress_chars[-1] * (progress_pos_in // self.n_block_share - progress_pos_out)
            if progress_pos_in < self.progress_len * self.n_block_share:
                progress_bar += self.progress_chars[progress_pos_in % self.n_block_share]
            progress_bar += self.progress_chars[0] * (self.progress_len - progress_pos_in // self.n_block_share - 1)
        else:
            progress_bar += self.progress_chars[-1] * (progress_pos_in // self.n_block_share)
            if progress_pos_out * self.n_block_share == progress_pos_in:
                if progress_pos_in < self.progress_len * self.n_block_share:
                    progress_bar += self.progress_chars[1]
            else:
                progress_bar += self.progress_chars[progress_pos_in % self.n_block_share]
                progress_bar += self.progress_chars[0] * (progress_pos_out - progress_pos_in // self.n_block_share - 1)
                progress_bar += self.progress_chars[1]
            progress_bar += self.progress_chars[0] * (self.progress_len - progress_pos_out - 1)
        first_not_act_loop = [i for i, x in enumerate(self.show_progresses + [None]) if not x][0]
        if sum(self.progresses[:first_not_act_loop]) == 0:
            print_time = datetime.now().strftime("%H:%M:%S")
            prefix = progress_bar + f"R--:--:--( -  s/it) |[{print_time}]"
        else:
            iter_estimations = []
            for i, t in enumerate(self.iter_estimations[first_not_act_loop-1::-1]):
                if t is not None: iter_estimations.append(t); continue
                if i == 0: iter_estimations.append(1); continue
                iter_estimations.append(iter_estimations[i - 1] * self.sizes[first_not_act_loop-i])
            iter_estimations = list(reversed(iter_estimations))
            t_iter = iter_estimations[-1]
            remaining_time = max(0, int(sum(t * (n - i - 1) for t, i, n, _ in zip(iter_estimations, self.progresses, self.sizes, range(first_not_act_loop))) + t_iter))
            secs = remaining_time % 60
            mins = (remaining_time // 60) % 60
            hours = remaining_time // 3600
            print_time = datetime.now().strftime("%H:%M:%S")
            prefix = progress_bar + f"R{hours:2d}:{mins:02d}:{secs:02d}({t_iter:.2f}s/it) |[{print_time}]"
        return prefix
    
    def __call__(self, arg:(list, int), n_timespan_store: int=None, show_progress_bar: bool=None, breakable: bool=None, break_key: str=None):
        if n_timespan_store is None: n_timespan_store = self.n_timespan_store
        if show_progress_bar is None: show_progress_bar = self.show_progress_bar
        if breakable is None: breakable = self.breakable
        if break_key is None: break_key = self.break_key
        i_loop = len(self.sizes)
        if isinstance(arg, int):
            self.sizes.append(arg)
            arg = range(arg)
        else:
            if not hasattr(arg, '__len__'):
                show_progress_bar = False
                self.sizes.append(-1)
                builtin_print(f"Warning: 'TaskProcess' cannot predict progress with non-list iterative object {arg}, consider casting the argument first or use show_progress_bar=False.")
            else: self.sizes.append(len(arg))
        self.show_progresses.append(show_progress_bar)
        self.progresses.append(0.)
        self.break_keys.append(break_key if breakable else None)
        self.should_break.append(False)

        start_in_this_loop = False
        if breakable:
            if not self.listener.is_listening(break_key):
                self.listener.add_listen(break_key)
            if not self.listener.started:
                start_in_this_loop = True
                self.listener.start()

        succeeded = False
        iter_timespans = []
        if len(self.iter_estimations) < i_loop + 1: self.iter_estimations.append(None)
        print(f"Loop starting at {datetime.now().strftime('%m/%d %H:%M:%S')}...")
        for i, x in enumerate(arg):
            iter_begin = datetime.now()
            self.progresses[i_loop] = i
            yield x
            iter_timespans.append((datetime.now() - iter_begin).total_seconds())
            if len(iter_timespans) > n_timespan_store: iter_timespans.pop(0)
            self.iter_estimations[i_loop] = median(iter_timespans)
            if breakable and self.listener.detected is not None and self.listener.detected in self.break_keys:
                b_loop = self.break_keys.index(self.listener.detected)
                self.listener.detected = None
                self.should_break = self.should_break[:b_loop] + [True] * (len(self.break_keys) - b_loop)
            if self.should_break[i_loop]: self.should_break[i_loop] = False; break
        else: succeeded = True
        if not succeeded: print("-- manual termination of the loop --")
        if start_in_this_loop: self.listener.stop()
        avouch(len(self.sizes) == i_loop + 1)
        self.sizes.pop(-1)
        self.progresses.pop(-1)
        self.show_progresses.pop(-1)
        if i_loop > 0: self.progresses[i_loop - 1] += 1

    def __enter__(self):
        self.parent_print = builtins.print
        builtins.print = self.print

    def __exit__(self):
        builtins.print = self.parent_print
    
iterate = TaskProgress()
        


try: from pycamia import info_manager
except ImportError: print("Warning: pycamia not loaded. ")

__info__ = info_manager(
    project = "PyCAMIA",
    package = "pyoverload",
    author = "Yuncheng Zhou", 
    create = "2023-09",
    fileinfo = "list all the dtypes. ",
    requires = ""
)

__all__ = """
    to_dtype
    to_rawdtype
    to_torch_type
    to_torch_dtype
    to_numpy_type
    to_numpy_dtype
""".split()

import re
from .typings import dtype

char_ = dtype('char')
byte_ = dtype('byte')
bool8 = bool_ = dtype('bool')
int_ = integer_ = dtype('int')
int8 = dtype('int8')
int16 = short_ = dtype('int16')
int32 = intc_ = dtype('int32')
int64 = intp_ = long_ = longlong_ = signedinteger_ = dtype('int64')
uint8 = ubyte_ = dtype('uint8')
uint16 = ushort_ = dtype('uint16')
uint32 = uintc_ = dtype('uint32')
uint64 = uint_ = uintp_ = ulonglong_ = unsignedinteger_ = dtype('uint64')
qint8 = dtype('qint8')
qint32 = dtype('qint32')
quint8 = dtype('quint8')
quint2x4 = dtype('quint2x4')
quint4x2 = dtype('quint4x2')
float_ = floating_ = dtype('float')
float16 = half_ = dtype('float16')
float32 = single_ = dtype('float32')
float64 = double_ = inexact_ = longdouble_ = longfloat_ = number_ = dtype('float64')
bfloat16 = dtype('bfloat16')
complex_ = dtype('complex')
complex32 = complexhalf_ = chalf_ = dtype('complex32')
complex64 = complexfloat_ = complexfloating_ = cfloat_ = csingle_ = singlecomplex_ = dtype('complex64')
complex128 = complexdouble_ = cdouble_ = clongfloat_ = clongdouble_ = longcomplex_ = dtype('complex128')
timedelta64 = dtype('<m8')

################################################################################
## The following chunk of code was used to generate the dtype names listed below.
################################################################################
## import torch
## import numpy
##
## torch_dtype = []
## torch_type = []
## numpy_type = []
##
## for n in dir(torch):
## 	   x = getattr(torch, n)
## 	   if isinstance(x, torch.dtype): torch_dtype.append(n)
## 	   if isinstance(x, type):
## 	   	   try:
## 			   torch.zeros(1).type(x)
## 			   torch_type.append(n)
## 		   except Exception: ...
##
## for n in dir(numpy):
## 	   x = getattr(numpy, n)
## 	   if isinstance(x, type) and issubclass(x, numpy.number): numpy_type.append(n)
################################################################################

def to_dtype(t):
    if t.__class__ == dtype: return t
    t_name = getattr(t, '__name__', str(t).split('.')[-1].split("'")[0])
    if t_name.endswith('Tensor'): t_name = t_name[:-len('Tensor')].lower()
    return globals().get(t_name, globals().get(t_name+'_', dtype(t)))

def to_rawdtype(t):
    t = to_dtype(t)
    t.module = None
    return t

torch_dtype = ['bfloat16', 'bool', 'cdouble', 'cfloat', 'chalf', 'complex128', 'complex32', 'complex64', 'double', 'float', 'float16', 'float32', 'float64', 'half', 'int', 'int16', 'int32', 'int64', 'int8', 'long', 'qint32', 'qint8', 'quint2x4', 'quint4x2', 'quint8', 'short', 'uint8']
torch_type = ['BFloat16Tensor', 'BoolTensor', 'ByteTensor', 'CharTensor', 'DoubleTensor', 'FloatTensor', 'HalfTensor', 'IntTensor', 'LongTensor', 'ShortTensor', 'Tensor', 'torch.ComplexHalfTensor', 'torch.ComplexFloatTensor', 'torch.ComplexDoubleTensor']
numpy_type = ['bool', 'bool8', 'bool_', 'byte', 'cdouble', 'cfloat', 'clongdouble', 'clongfloat', 'complex128', 'complex64', 'complex_', 'complexfloating', 'csingle', 'double', 'float16', 'float32', 'float64', 'float_', 'floating', 'half', 'inexact', 'int16', 'int32', 'int64', 'int8', 'int_', 'intc', 'integer', 'intp', 'longcomplex', 'longdouble', 'longfloat', 'longlong', 'number', 'short', 'signedinteger', 'single', 'singlecomplex', 'timedelta64', 'ubyte', 'uint', 'uint16', 'uint32', 'uint64', 'uint8', 'uintc', 'uintp', 'ulonglong', 'unsignedinteger', 'ushort']
torch_dtype_name = {to_rawdtype(n): n for n in torch_dtype}
torch_type_name = {to_rawdtype(n): n for n in torch_type}
numpy_type_name = {to_rawdtype(n): n for n in numpy_type}

to_torch_dtype_special = {'byte': 'uint8', 'char': 'int8', '<m8': 'int64'}

def to_torch_dtype(t):
    import torch
    t = to_dtype(t)
    t.module = None
    full_name = f"{t.name}{t.bits}"
    if full_name in to_torch_dtype_special:
        t.name, t.bits = re.findall(r"([<a-zA-Z]+)([x0-9]*)", to_torch_dtype_special[full_name])[0]
    if not full_name: return torch.float32
    if t.name.startswith('u'): t.name = t.name[1:]
    type_name = torch_dtype_name.get(t, None)
    if type_name is None: raise TypeError(f"Unrecognized dtype {t} for torch dtypes.")
    return getattr(torch, type_name)

def to_torch_type(t):
    import torch
    t = to_dtype(t)
    t.module = None
    if t.name.startswith('q'): t.name = t.name[1:]
    if t.name.startswith('u'): t.name = t.name[1:]
    if t.name == '<m': t.name = 'int'; t.bits = 64
    if t.bits in ('2x4', '4x2'): t.bits = 8
    type_name = torch_type_name.get(t, torch_type_name.get(dtype(t.name), None))
    if type_name is None: raise TypeError(f"Unrecognized dtype {t} for torch types.")
    return getattr(torch, type_name, type_name)

to_numpy_dtype_special = {'bfloat16': 'float16', 'complex32': 'complex64', '': 'float32'}

def to_numpy_dtype(t):
    import numpy
    t = to_dtype(t)
    full_name = f"{t.name}{t.bits}"
    return numpy.dtype(to_numpy_dtype_special.get(full_name, full_name))

to_numpy_special = {'bfloat16': 'float16', 'complex32': 'complex64', 'bool': 'uint8', 'char': 'int8'}

def to_numpy_type(t):
    import numpy
    t = to_dtype(t)
    t.module = None
    if t.name.startswith('q'): t.name = t.name[1:]
    full_name = f"{t.name}{t.bits}"
    if full_name in to_numpy_special:
        t.name, t.bits = re.findall(r"([<a-zA-Z]+)([x0-9]*)", to_numpy_special[full_name])[0]
    if not full_name: return numpy.float32
    if t.bits and isinstance(t.bits, str): t.bits = eval(t.bits.replace('x', '*'))
    type_name = numpy_type_name.get(t, None)
    if type_name is None: raise TypeError(f"Unrecognized dtype {t} for numpy types.")
    return getattr(numpy, type_name)




from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "batorch",
    fileinfo = "Tensor functions for `batorch`.",
    requires = ["numpy", "torch", "pycamia", "pyoverload", "matplotlib", "sympy"]
)

__all__ = """
    batorch_wrapper
    conv
    crop_as
    decimal
    divide
    dot
    equals
    gaussian_kernel
    image_grid               imagegrid
    mask_center
    one_hot
    pad
    permute_space
    normalize01
    input_shape

    grad_image               Jacobian
    up_scale                 down_scale
    
    upper_matrix             upper_triangular_matrix
    lower_matrix             lower_triangular_matrix
    skew_symmetric           cross_matrix
    skew_symmetric_params    uncross_matrix
    
    norm                     norm2
    Fnorm                    Fnorm2
    frobenius_norm
    meannorm                 meannorm2
    mean_norm                mean_norm2
    
    summary                  display
    num_model_params
""".split()

import math, re

with __info__:
    import numpy as np
    import torch
    import batorch as bt
    from functools import wraps
    from sympy import symbols, Eq, solve
    from pyoverload import *
    from pycamia import avouch, touch, alias, decorator
    from pycamia import get_environ_vars, get_snakenames, get_args_expression
    from pycamia import SPrint, Version, argmin, tokenize, to_tuple, ByteSize
    from matplotlib import pyplot as plt
    
@decorator
def batorch_wrapper(func: callable=None, **options):
    """
    The wrapper for batorch functions, ensuring the inputs of the function are batorch 
        tensors, and the return type will be casted as the input type for non-batorch inputs. 
    """
    if func is None: return lambda f: batorch_wrapper(f, **options)
    i_arg = options.pop('i_arg', None)
    default_aux = options
    @wraps(func)
    def wrapper_func(*args, **kwargs):
        values = get_arg_values(get_func_info(func), *args, **kwargs)
        f_varnames = func.__code__.co_varnames
        new_args = []
        new_kwargs = {}
        input_class = None
        for i, (n, v) in enumerate(zip(f_varnames, values)):
            if func.__annotations__.get(n, None) in (array, bt.Tensor):
                if i_arg is None and input_class is None or i == i_arg: input_class = v.__class__
                if v is None: ...
                elif not isinstance(v, torch.Tensor): v = bt.tensor(v)
                elif not isinstance(v, bt.Tensor): v = v.as_subclass(bt.Tensor).init_special()
            if (func.__code__.co_flags & 0x04) and i == func.__code__.co_argcount + func.__code__.co_kwonlyargcount: new_args.extend(v)
            elif (func.__code__.co_flags & 0x08) and i == func.__code__.co_argcount + func.__code__.co_kwonlyargcount + ((func.__code__.co_flags & 0x04) >> 2): new_kwargs.update(v)
            elif i < len(args): new_args.append(v)
            else: new_kwargs[n] = v
        ret = func(*new_args, **new_kwargs)
        if not isinstance(ret, bt.Tensor): return ret
        ret_obj = None
        if input_class is not None:
            if issubclass(input_class, torch.Tensor) and input_class != bt.Tensor: ret_obj = ret.as_subclass(input_class)
            elif issubclass(input_class, np.ndarray): ret_obj = ret.detach().cpu().numpy()
            elif issubclass(input_class, list): ret_obj = ret.detach().cpu().numpy().tolist()
            elif issubclass(input_class, tuple): ret_obj = tuple(ret.detach().cpu().numpy().tolist())
        if ret_obj is None: ret_obj = ret
        for k, default in default_aux.items():
            aux = getattr(ret, k, default)
            if issubclass(default.__class__, torch.Tensor):
                if input_class != bt.Tensor: aux = (aux if isinstance(aux, torch.Tensor) else torch.tensor(aux)).as_subclass(default.__class__)
                else: aux = aux if isinstance(aux, bt.Tensor) else bt.tensor(aux)
            elif issubclass(default.__class__, np.ndarray): aux = aux.detach().cpu().numpy() if isinstance(aux, torch.Tensor) else np.array(aux)
            elif issubclass(default.__class__, list): aux = aux.detach().cpu().numpy().tolist() if isinstance(aux, torch.Tensor) else list(aux)
            elif issubclass(default.__class__, tuple): aux = tuple(aux.detach().cpu().numpy().tolist()) if isinstance(aux, torch.Tensor) else tuple(aux)
            try: setattr(ret_obj, k, aux)
            except AttributeError: pass
        return ret_obj
    return wrapper_func

# @overload
# def batorch_wrapper(func: callable, i_arg=None, **default_aux):
#     @wraps
#     def wrapper_func(*args, **kwargs):
#         values = get_arg_values(get_func_info(func), *args, **kwargs)
#         f_varnames = func.__code__.co_varnames
#         n_posarg = func.__code__.co_posonlyargcount
#         new_args = []
#         new_kwargs = {}
#         input_class = None
#         for i, (n, v) in enumerate(zip(f_varnames, values)):
#             if func.__annotations__.get(n, None) == bt.Tensor:
#                 if i_arg is None and input_class is None or i == i_arg: input_class = v.__class__
#                 if not isinstance(v, torch.Tensor): v = bt.tensor(v)
#                 else: v = v.as_subclass(bt.Tensor)
#             if i < n_posarg: new_args.append(v)
#             else: new_kwargs[n] = v
#         ret = func(*new_args, **new_kwargs)
#         if input_class is not None:
#             if issubclass(input_class, torch.Tensor): ret_obj = ret.as_subclass(input_class)
#             elif issubclass(input_class, np.ndarray): ret_obj = ret.detach().cpu().numpy()
#             elif issubclass(input_class, list): ret_obj = ret.detach().cpu().numpy().tolist()
#             elif issubclass(input_class, tuple): ret_obj = tuple(ret.detach().cpu().numpy().tolist())
#         for k, default in default_aux:
#             aux = getattr(ret, k, default)
#             if issubclass(default.__class__, torch.Tensor): aux = (aux if isinstance(aux, torch.Tensor) else torch.tensor(aux)).as_subclass(default.__class__)
#             elif issubclass(default.__class__, np.ndarray): aux = aux.detach().cpu().numpy() if isinstance(aux, torch.Tensor) else np.array(aux)
#             elif issubclass(default.__class__, list): aux = aux.detach().cpu().numpy().tolist() if isinstance(aux, torch.Tensor) else list(aux)
#             elif issubclass(default.__class__, tuple): aux = tuple(aux.detach().cpu().numpy().tolist()) if isinstance(aux, torch.Tensor) else tuple(aux)
#             setattr(ret_obj, k, aux)
#         return ret_obj
#     return wrapper_func

# @overload
# def batorch_wrapper(i_arg: int | null =None, **default_aux):
#     return wraps(lambda f: batorch_wrapper(f, i_arg, **default_aux))

@batorch_wrapper
def conv(tensor: bt.Tensor, kernel: bt.Tensor, padding='SAME'):
    """
    Convolution for a batorch tensor: only in space dimensions. 
    
    Args:
        tensor (bt.Tensor): With shape ({n_batch: optional}, [n_channel_in: optional], *n_space).
        kernel (bt.Tensor): With shape ([n_channel_out: optional], [n_channel_in: optional], *n_kernel_size).
        padding (str, int, tuple): Convolution padding: string 'SAME' stands for not changing the size of 'tensor'. 
    
    Returns:
        output (bt.Tensor): With shape ({n_batch: optional}, [n_channel_out: optional], *(n_space - n_kernel_size // 2)).
    """
    input_shape = tensor.shape
    if not tensor.has_batch: tensor = tensor.unsqueeze({})
    if not tensor.has_feature: tensor = tensor.unsqueeze([])
    n_dim = tensor.n_space_dim
    avouch(kernel.n_space_dim == n_dim, f"Kernel for {n_dim}D convolution should be of size ([channel_out, channel_in], n_kernel_1, ..., n_kernel_{n_dim}) with a dimension of {n_dim}. ")
    if kernel.has_feature:
        avouch(kernel.n_feature_dim == 2, f"Kernel for {n_dim}D convolution should have exactly 2 channel dimensions, only size ([channel_out, channel_in], n_kernel_1, ..., n_kernel_{n_dim}) is allowed, not {kernel.shape}.")
        avouch(tensor.has_channel, f"Tensor input for convolution should have channel dimension (one feature dimension): not tensor size {tensor.shape} and kernel size {kernel.shape}. Use kernel without special dimension to perform spatial-dimension-only convolution. ")
        avouch(tensor.n_feature_dim == 1 and tensor.n_channel == kernel.feature[1], f"Channels of convolution should match: tensor size {str(tensor.shape).replace('[', '[>').replace(']', '<]')} and kernel size {str(kernel.shape).replace(f'{kernel.feature[-1]}]', f'>{kernel.feature[-1]}<]')}. ")
        space_conv = False
    else: space_conv = True

    is_same = False
    if isinstance(padding, str) and padding == 'SAME':
        is_same = True
        padding = tuple(x // 2 for x in kernel.space)
    if isinstance(padding, int):
        padding = (padding,) * n_dim
    avouch(isinstance(padding, tuple), f"Unrecognized argument 'padding' for convolution: {padding}. ")
    conv = eval("bt.nn.functional.conv%dd"%n_dim)
    
    if space_conv:
        prev_batch_feature = bt.Size({tensor.n_batch}) + tensor.feature
        tensor = tensor.merge_dims([], {}).unsqueeze([])
        kernel = kernel.unsqueeze(0, 0)
    result = conv(tensor, kernel.type(tensor.dtype), padding=padding)
    if space_conv: result = result.squeeze([]).splitdim({}, prev_batch_feature)
    if not input_shape.has_batch: result.squeeze_({})
    if not input_shape.has_feature: result.squeeze_([])
    if is_same: result = crop_as(result, input_shape.space)
    
    # if kernel.size(0) > 1:
    #     if input_shape.has_channel:
    #         input_shape = input_shape.with_dim_size(input_shape.channel_dim, kernel.size(0))
    
    # if is_same: result = crop_as(result, tensor).view(*input_shape)
    # else: result = result.view(input_shape.with_space(tuple(t - k + 1 + 2 * p for t, k, p in zip(tensor.space, kernel.space, padding))))
    return result

@alias("meannorm2", "mean_norm2", root = False, mean = True)
@alias("meannorm", "mean_norm", mean = True)
@alias("Fnorm2", "norm2", root = False)
@alias("Fnorm", "frobenius_norm")
@batorch_wrapper
def norm(tensor: bt.Tensor, p = 2, root = True, mean = False, dim=null):
    """
    The norm of a tensor in (the first available condition):
    (1) all feature dimensions (if n_feature_dim >= 1);
    (2) all space dimensions (if n_space_dim >= 1); 
    (3) all sequence dimensions (if n_sequence_dim >= 1). 
    Use 'dim=None' to compute for all dimensions. 

    Args:
        tensor (bt.Tensor): The tensor to compute norm, commonly in shape ({n_batch}, [n_channel], n_1, ..., n_{n_dim})
        p (int, optional): The order of norm. Defaults to 2.
            p = 1 for L1-norm/ LASSO;
            p = 2 for L2-norm or Frobenius;
            p = math.inf / float('inf') for L∞-norm / maximum. 
        root (bool, optional): Whether to calculate the root. Defaults to True for common norms.
        mean (bool, optional): Whether to use mean across the dimensions. Defaults to False for common norms.
        dim (_type_, optional): The dimension to perform norm. Defaults to null, which auto selects the dimensions.
            One need to manually set 'dim=None' to perform norm in all dimensions. 
    """
    if dim is None: dim = bt.exist_dim(tensor, dim)
    else: dim = bt.linalg_dim[1:](tensor, None if dim is null else dim)
    tensor = tensor.abs()
    if p == float('inf'): return tensor.max(*dim)
    if mean: return (tensor ** p).mean(*dim) ** (1 / p) if root else (tensor ** p).mean(*dim)
    else: return (tensor ** p).sum(*dim) ** (1 / p) if root else (tensor ** p).sum(*dim)

@batorch_wrapper
def decimal(tensor: bt.Tensor):
    """
    The decimal part of tensor: {x} = x-[x]. 
    """
    return tensor - bt.floor(tensor)

@batorch_wrapper
def divide(a: bt.Tensor, b: bt.Tensor, limit=1, tol=1e-6):
    """
    def divide(a: bt.Tensor, b: bt.Tensor, limit=1, tol=1e-6)
    Element-wise division `a / b`, except it equals to `limit` if the divider `b` is smaller than `tol` (which is regarded as the division of zero)
    """
    a_s, b_s = a.shape ^ b.shape
    a = a.view(a_s)
    b = b.view(b_s)
    shape = bt.Size(max(x, y) for x, y in zip(a_s, b_s)).special_from(a_s)
    return bt.where(b.abs() < tol, bt.where(a.abs() < tol, bt.zeros(shape), limit * bt.ones(shape)), a / bt.where(b.abs() < tol, tol * bt.ones(shape), b))

@batorch_wrapper
def equals(x: bt.Tensor, y: bt.Tensor, tol=1e-6):
    """
    Determine whether the corresponding elements in two tensors are equal to each other. 
    The region of equivalence is a union of a diamond area at the origin point (⟐) 
        and a cross area in shape 'x' (with borders not passing the origin: '⪥').
    The cross area includes points near lines y=x and y=-x with an angle dilation of approximately arctan(tol). 
    The diamond area has a side length of √2*tol. 
    The reason why a diamond area is added at the origin instead of square (⊡) or circular region (⊙) is that,
        for inputs of difference storage accuracy (e.g. int & float), higher tolerance should be placed for errors 
        in the lower accuracy (e.g. float) tensor than in the higher accuracy (e.g. int) tensor, causing peaks in tolerance map. 
    """
    return ((bt.abs(x - y) / (bt.abs(x) + bt.abs(y) + tol)) < tol) | (bt.abs(x) + bt.abs(y) < tol)

def gaussian_kernel(n_dim = 2, kernel_size = 3, sigma = 0, normalize = True):
    r"""
    Create a bt.Tensor Gaussian kernel of size (kernel_size, ..., kernel_size). 
                                                \---{n_dim} kernel_size's---/

    Args:
        n_dim (int, optional): The dimension of the kernel. Defaults to 2.
        kernel_size (int, optional): The side length of the kernel. Defaults to 3.
        sigma (int or float, optional): The standard deviation sigma of Gaussian distribution, 
                                            in unit 'pixel'. Defaults to 0.
        normalize (bool, optional): Whether the sum of kernel should be 1. Defaults to True.
    """
    if not isinstance(kernel_size, bt.torch.Tensor):
        if not isinstance(kernel_size, (tuple, list)): kernel_size = (kernel_size,)
        kernel_size = list(kernel_size)
        if len(kernel_size) == 1: kernel_size *= n_dim
        kernel_size = bt.Tensor(kernel_size)
    elif not isinstance(kernel_size, bt.Tensor): kernel_size = kernel_size.as_subclass(bt.Tensor).init_special()
    kernel_size.n_feature_dim = 1
    radius = (kernel_size - 1) / 2

    if sigma == 0: sigma = radius * 0.6
    elif not isinstance(sigma, bt.torch.Tensor):
        if not isinstance(sigma, (tuple, list)): sigma = (sigma,)
        sigma = list(sigma)
        if len(sigma) == 1: sigma *= n_dim
        sigma = bt.Tensor(sigma)
    elif not isinstance(sigma, bt.Tensor): sigma = sigma.as_subclass(bt.Tensor).init_special()
    sigma.n_feature_dim = 1

    grid = bt.image_grid(*kernel_size.tolist()).float()
    kernel = bt.exp(- (((grid - radius) / sigma) ** 2).sum(0) / 2)
    return (kernel / kernel.sum()) if normalize else kernel

@batorch_wrapper
def dot(g1: bt.Tensor, g2: bt.Tensor, dim=null):
    """
    The dot product of two tensors in (the first available condition):
    (1) all feature dimensions (if n_feature_dim >= 1);
    (2) all space dimensions (if n_space_dim >= 1); 
    (3) all sequence dimensions (if n_sequence_dim >= 1). 
    Use 'dim=None' to compute for all dimensions. 
    """
    if dim is None: dim = bt.exist_dim(g1, dim)
    else: dim = bt.linalg_dim[1:](g1, None if dim is null else dim)
    avouch(g1.shape == g2.shape, "Please make sure the dot product recieve two tensor of a same shape. Use 'bt.expand_to' to adjust if necessary. ")
    avouch(g1.n_feature_dim == g2.n_feature_dim, "Please make sure the inputs of 'dot' have a same feature dimension. ")
    return (g1 * g2).sum(*dim)

@batorch_wrapper
def Jacobian(X, Y, dt=1, pad=False):
    """
    The Jacobian matrix; Note that it is a transpose of grad_image if X is standard grid as it follows the orientation of Jacobian.
    
    Args:
        X (bt.Tensor): in shape ({n_batch}, [n_dim], n_1, ..., n_{n_dim})
        Y (bt.Tensor): in shape ({n_batch}, [n_feature], n_1, ..., n_{n_dim})
        
    Returns:
        Jacobian(output, pad = True): in shape ({n_batch}, [n_feature, n_dim], n_1, ..., n_{n_dim})
        Jacobian(output, pad = False): in shape ({n_batch}, [n_feature, n_dim], n_1 - dt, ..., n_{n_dim} - dt)
    """
    dX = grad_image(X, dx=dt, pad=pad) # ({n_batch}, [n_dim, n_dim], n_1, ..., n_{n_dim})
    dY = grad_image(Y, dx=dt, pad=pad) # ({n_batch}, [n_dim, n_feature], n_1, ..., n_{n_dim})
    # dy_k/dx_i = sum_j [(dy_k/dt_j) / (dx_i/dt_j)] => Jac[{b}, [k, i], ...] = sum_j Jac_j[{b}, [j, k, i], ...] = dY[{b}, [j, k, *], ...] / dX[{b}, [j, *, i], ...]
    return bt.divide(dY.unsqueeze([3]), dX.unsqueeze([2]), 0.0, tol=1e-4).sum([0])

@batorch_wrapper
def grad_image(array: bt.Tensor, dx=1, pad=False):
    '''
        Gradient image / tensor of array (dx is the displacement for finite difference). 
        If pad is not True, the image will trim. The sizes should be: 
        array: ({n_batch}, [n_feature], n_1, ..., n_{n_dim})
        output (pad = True): ({n_batch}, [n_dim, n_feature], n_1, ..., n_{n_dim})
        output (pad = False): ({n_batch}, [n_dim, n_feature], n_1 - dx, ..., n_{n_dim} - dx)
        OR:
        array: ({n_batch}, n_1, ..., n_{n_dim})
        output (pad = True): ({n_batch}, [n_dim], n_1, ..., n_{n_dim})
        output (pad = False): ({n_batch}, [n_dim], n_1 - dx, ..., n_{n_dim} - dx)
    '''
    if not array.has_batch: array = array.unsqueeze({})
    output = []
    size = array.space
    if not pad: size = tuple(x - dx for x in size)
    for d in range(*array.space_range):
        b = (slice(None, None),) * d + (slice(dx, None),) + (slice(None, None),) * (array.ndim - d - 1)
        a = (slice(None, None),) * d + (slice(None, -dx),) + (slice(None, None),) * (array.ndim - d - 1)
        output.append(bt.crop_as((array[b] - array[a]) / dx, size))
    return bt.stack(output, [int(array.has_batch)])

@overload
@batorch_wrapper(roi=tuple())
def crop_as(x: array, y: tuple, center: tuple, fill: number=0) -> array:
    """
    crop a Tensor `x` as the shape given by `y`. 

    Args:
        x (bt.Tensor): The data to crop (or pad if the target shape is larger), 
            Shape: ({n_batch}, [*n_feature], *n_space). 
            Note that only the space dimensions of the tensor are cropped/padded by default.
        y (bt.Tensor or tuple): The target shape in tuple or another tensor to provide the target shape.
        center (tuple, optional): The center of the target box. Defaults to the center of x's shape. 
            Note: Do calculate the center w.r.t. input `x` if one is expanding the tensor, 
                as `x`'s center coordinates in y-space is different from `y`'s center 
                coordinates in x-space (which is correct).
        fill (number, optional): The number to fill for paddings. Defaults to 0.
    """
    size_x = x.shape
    size_y = bt.Size(y)

    if isinstance(size_y, bt.Size) and size_x.n_space_dim == size_y.n_space_dim:
        size_y = size_y.space
    exp_size_x, exp_size_y = size_x ^ bt.Size(-1 if s == 1 else s for s in size_y).special_from(size_y)
    avouch(len(exp_size_x) == len(size_x), f"Cannot match crop size {size_y} with tensor size {size_x} in 'crop_as'. ")
    exp_size_y = bt.Size(-s if s in (1, -1) else s for s in exp_size_y).special_from(exp_size_y)
    size_x, size_y = exp_size_x, exp_size_y

    center = bt.Size(center)
    if len(center) == len(size_x): center = center.special_from(size_x)
    elif len(center) == size_x.n_space_dim: center = (center ^ size_x)[0]
    elif len(x for x in center if x >= 0) == len(x for x in size_y if x >= 0):
        center = bt.Size(a if b >= 0 else -1 for a, b in zip(center, size_y)).special_from(size_x)
    avouch(len(center) == len(size_x), TypeError("Mismatch dimensions for the center in 'crop_as', please use -1 if the dimension that is centered or doesn't need cropping. "))
    center = bt.Size(-1 if sx == sy or sy < 0 else (sx / 2 if c < 0 else c) for c, sx, sy in zip(center, size_x, size_y)).special_from(center)
    size_y = bt.Size(sx if sy < 0 else sy for sx, sy in zip(size_x, size_y)).special_from(size_y)

    z = fill * bt.ones(*size_y).to(x.device).type_as(x)
    """
    We then calculate the intersected boxes in source 'x' and target 'z', 
    to manage the occasions when the two regions are not containing each other (like: ⧉). 
    """
    def intersect(u, v): return max(u[0], v[0]), min(u[1], v[1])
    z_box = [(0, lx) if m < 0 or ly < 0 else intersect((0, ly), (- round(float(m - float(ly) / 2)), - round(float(m - float(ly) / 2)) + lx)) for m, lx, ly in zip(center, size_x, size_y)]
    x_box = [(0, lx) if m < 0 or ly < 0 else intersect((0, lx), (+ round(float(m - float(ly) / 2)), + round(float(m - float(ly) / 2)) + ly)) for m, lx, ly in zip(center, size_x, size_y)]
    # if the two boxes are seperated
    if any([r[0] >= r[1] for r in z_box]) or any([r[0] >= r[1] for r in x_box]): z.roi = None; return z
    region_z = tuple(slice(u, v) for u, v in z_box)
    region_x = tuple(slice(u, v) for u, v in x_box)
    z[region_z] = x[region_x]
    z.roi = region_x
    z.special_from(size_x)
    return z

@overload
def crop_as(x: array, y: array, center: tuple, fill: number=0) -> array:
    return crop_as(x, y.shape, center, fill)

@overload
def crop_as(x: array, y: union(tuple, array), fill: number=0) -> array:
    center = tuple(-1 for m in x.shape)
    return crop_as(x, y, center, fill)

@overload
def crop_as(x: array, *y: int) -> array:
    center = tuple(-1 for m in x.shape)
    return crop_as(x, y, center)

@batorch_wrapper
def pad(x: bt.Tensor, p = 1, fill = 0):
    """
    Pad the tensor by `p` pixels with values `fill`. 
    p can be a tuple of len `n_dim` to pad differently for each dimension. 
    """
    p = to_tuple(p)
    if len(p) == 1: p *= x.n_space_dim
    return crop_as(x, tuple(s + 2 * q for s, q in zip(x.space, p)), fill = fill)

def add_special(size, special, fill=1):
    s = special
    if len(s) == 0: pass
    elif len(s) == 1: size = size[:s[0]] + (fill,) + size[s[0]:]
    else: size = size[:s[0]] + (fill,) + size[s[0]:s[1]-1] + (fill,) + size[s[1]-1:]
    return size

@batorch_wrapper
def up_scale(image: bt.Tensor, *scaling:int):
    """
    up_scale(tensor, s_1, ..., s_{n_dim}): equivalent to tensor.amplify(s_i, i)
    """
    if len(scaling) == 0:
        scaling = (1,)
    elif len(scaling) == 1 and iterable(scaling[0]):
        scaling = scaling[0]
    if len(scaling) == 1:
        if isinstance(scaling[0], int):
            scaling *= image.n_space_dim
            scaling = add_special(scaling, image.special_dims, 1)
        else: raise TypeError("Unknown scaling type for 'up_scale'. ")
    elif len(scaling) < image.n_dim and len(scaling) == image.n_space_dim:
        scaling = add_special(scaling, image.special_dims, 1)
    for i, s in enumerate(scaling):
        image = (
            image
            .transpose(i, -1)
            .unsqueeze(-1)
            .repeat((1,) * image.ndim + (int(s),))
            .flatten(-2)
            .transpose(i, -1)
        )
    return image

@batorch_wrapper
def down_scale(image: bt.Tensor, *scaling:int):
    """
    down_scale(tensor, s_1, ..., s_{n_dim}): equivalent to tensor[::s_i, ..., ::s_{n_dim}]
    """
    if len(scaling) == 0:
        scaling = (1,)
    elif len(scaling) == 1 and iterable(scaling[0]):
        scaling = scaling[0]
    if len(scaling) == 1:
        if isinstance(scaling[0], int):
            scaling *= image.n_space_dim
            scaling = add_special(scaling, image.special_dims, 1)
        else: raise TypeError("Unknown scaling type for 'down_scale'. ")
    elif len(scaling) < image.n_dim and len(scaling) == image.n_space_dim:
        scaling = add_special(scaling, image.special_dims, 1)
    return image[tuple(slice(None, None, s) for s in scaling)]

@alias("imagegrid")
def image_grid(*shape, device=bt.default_device()):
    """
    Create the indices for a space represented by tuple, or the space-dimension space of a tensor. 
    Note that the indices are distributed as indexing = 'ij' in the representation of keyword argument for torch.meshgrid.
    
    Outputs:
        (1) For input of space shape: (n_1, n_2, ..., n_{n_dim}), 
            output a Tensor in shape ([n_dim], n_1, n_2, ..., n_{n_dim}).
        (2) For input of tensor shape: ({n_batch}, n_1, ..., n_{n_dim}, 'n_sequence'), 
            output a Tensor in shape ([n_dim], n_1, ..., n_{n_dim}). 
        (3) ** WARNING: Please be careful with this usage as the feature dimension of the output is irrelevant to the input. **
            For input of tensor shape with feature: ({n_batch}, [n_feature], n_1, ..., n_{n_dim}, 'n_sequence'), 
            output a Tensor in shape ([n_dim], n_1, ..., n_{n_dim}). 
    """
    if len(shape) == 1 and isinstance(shape[0], (list, tuple)): shape = shape[0]
    if len(shape) == 1 and hasattr(shape[0], 'space'): shape = shape[0].space
    if len(shape) == 1 and hasattr(shape[0], 'shape'): shape = shape[0].shape
    shape = tuple(bt.Size(*shape).space)
    kwargs = {'indexing': 'ij'} if bt.__torchversion__ >= Version('1.10') else {}
    if any(x < 0 for x in shape): raise TypeError(f"Negative size in shape {shape} for image grid.")
    if len(shape) == 0: return bt.zeros([0])
    return bt.stack(bt.meshgrid(*[bt.arange(x, device=device) for x in shape], **kwargs), [])

@batorch_wrapper
def mask_center(mask: bt.Tensor):
    """
    Compute the center of gravity for a mask.
    """
    avouch(issubclass(mask.dtype, dtype(bool)), TypeError("Only bool mask is allowed in function 'bt.mask_center'. "))
    coords = bt.image_grid(mask).change_special_dim([], bt.Size(0).with_func_dim(True))
    return (coords * mask).float().sum(...) / mask.sum(...)

# def linear(input, weight, bias):
#     result = input @ weight.T
#     if bias is not None:
#         if bias.dim() == 2:
#             return result + bias
#         return result + bias.unsqueeze(0)
#     return result

@batorch_wrapper
def one_hot(k: bt.Tensor, num_classes=None):
    """
    one_hot(k, n) -> bt.Tensor
    Create a one-hot vector as a `bt.Tensor` with length `n` with all elements 0 except the `k`-th element being 1. 
    e.g. k == -1 gives tensor([0, 0, ..., 0, 1])
    """
    avouch(not k.is_floating_point(), "'one_hot' requires the index argument of integers. ")
    if num_classes is None: num_classes = k.max().item() + 1
    l = bt.zeros(k.shape + bt.func_dim_size(num_classes)).int(); l[*image_grid(*k.shape), k] = 1
    return l

@batorch_wrapper
def permute_space(data: bt.Tensor, *dims):
    """
    permute the space section in data. len(dims) should be data.n_space_dim. 
    """
    if len(dims) == 1 and isinstance(dims[0], (list, tuple)): dims = dims[0]
    avouch(len(dims) == data.n_space_dim, TypeError(f"'permute_space' only takes 'dims' of length data.n_space_dim={data.n_space_dim}, currently {dims}."))
    dimensions = list(range(data.space_start)) + [d + data.space_start for d in dims] + list(range(data.space_stop, data.n_dim))
    return data.permute(*dimensions)

@batorch_wrapper
def normalize01(data: bt.Tensor, func='linear'):
    """
    Normalize the data to range [0, 1]. Outliers are automatically excluded. 
        The result is continuous, even at the two ends.

    Args:
        data (bt.Tensor): The input data of size ({n_batch}, [n_feature], n_1, ..., n_r) [r=n_dim]
        func (str): The basic mapping function. Defaults to 'linear'.
            'linear': x => (x - x_min) / (x_max - x_min)
            'sigmoid': x => 1 / (1 + 19 ^ (2 * x - x_max - x_min) / (x_max - x_min))
    """
    if data.has_feature:
        norm = (data ** 2).sum([], keepdim=True).clamp(bt.eps).sqrt()
        m = norm.quantile(0.05, ..., keepdim=True)
        M = norm.quantile(0.95, ..., keepdim=True)
        new_norm = normalize01(norm.squeeze([]), func=func).view_as(norm)
        non_zero_norm = bt.where(norm == 0, bt.ones_like(norm), norm)
        return data * bt.where(new_norm == 0, bt.ones_like(new_norm), new_norm / non_zero_norm)
    else:
        m = data.quantile(0.05, ..., keepdim=True)
        M = data.quantile(0.95, ..., keepdim=True)
        clamped = bt.min(bt.max(data, m), M)
        if M.max().item() < 1.1 and m.min().item() > -0.1: return clamped.clamp(0, 1)
        if func.lower() == 'linear':
            return (clamped - m) / (M - m).clamp(bt.eps)
        elif func.lower() == 'sigmoid':
            return 1 / (bt.exp(- math.log(19) * (2 * clamped - M - m) / (M - m).clamp(bt.eps)) + 1)
        else: raise TypeError("Only 'linear' and 'sigmoid' functions are available for argument 'func' in 'normalize01'. ")

class input_shape:
    
    def __init__(self, **kwargs):
        self.vars = get_environ_vars()
        self.default_values = kwargs
        self.exp_val_pair = []
        self.opt_exp_val_pair = []
        self.index = 0
        
    def get_variables(self, expr):
        variables = []
        while True:
            try: eval(expr); break
            except NameError as e:
                name = e.__str__().split("'")[1]
                variables.append(name)
                exec(f"{name} = 2")
        return variables
    
    def declose(self, expr):
        for closure in ('{}', '[]', '""', "''", '()'):
            if expr.startswith(closure[0]) and expr.endswith(closure[1]):
                return expr.strip(closure)
        return expr
    
    def split(self, shape_expr, sep=','):
        return [x.strip() for x in tokenize(shape_expr, sep=sep) if x.strip()]
    
    def record_expr_value(self, expr, value):
        if isinstance(expr, str): expr = [expr]
        if isinstance(value, int): value = [value]
        for e, v in zip(expr, value):
            if v > 1: self.exp_val_pair.append((e, v))
            else: self.opt_exp_val_pair.append((e, v))
    
    def set(self, **kwargs):
        """
        set shape information of tensors for a function.
        
        Args:
            kwargs: data_name = shape pairs. 
            data_name (str): the name of local variable. 
            shape (str): a string representing the shape of data. The expression is in the shape format. 
                The unknown number of dimensions can be specified with '...' and variable can be specified with words. 
        
        Examples::
            >>> data = bt.zeros({4}, [3, 3], 24, 25, 36)
            >>> data2 = bt.zeros({4}, [3])
            >>> input_shape(n_dim=2).set(
            ...     data = "({n_batch}, [n_dim+1, n_dim+1], n_1, ..., n_r) [r=n_dim]", 
            ...     data2 = "({n_batch}, [(n_dim+1) * n_dim / 2])"
            ... )
            ...
            >>> print(n_batch, n_dim, r)
            4 2 3
        """
        self.exp_val_pair = []
        self.opt_exp_val_pair = []
        variable_names = []
        const_eqs = []
        n_omits = {}
        for data_name, shape_expr in kwargs.items():
            data = self.vars[data_name]
            
            # get constraint eqs
            shape_expr, *const_eq_strs = self.split(shape_expr, sep='')
            is_optional = False
            for ceq in const_eq_strs:
                ceq.strip('[]')
                for x in ceq.split(';'):
                    x = x.strip().strip('[]')
                    if x.lower() == 'optional': is_optional = True; continue
                    left, right = x.split('=')
                    variable_names.extend(self.get_variables(left))
                    variable_names.extend(self.get_variables(right))
                    const_eqs.append(left + '-' + right)
            avouch(isinstance(data, bt.Tensor) or data is None and is_optional,
                   TypeError(f"'{data_name}' needs to be a batorch tensor of shape {shape_expr}, instead of type {data.__class__.__name__}"))
            if data is None: continue
                    
            # get variable names
            variable_names.extend(self.get_variables(shape_expr))
            
            # seperate shape terms
            shape_terms = self.split(shape_expr.strip('()'))
            element_terms = []
            ref_shape = bt.Size()
            opt_shape = bt.Size()
            for term in shape_terms:
                eles = self.split(self.declose(term))
                optional_terms = [x for x in eles if x.endswith(':optional') or x=='n_batch']
                if len(optional_terms) > 0: opt_shape += bt.Size(eval(term[0] + '-1' + term[-1]))
                eles = [x[:-9] if x.endswith(':optional') else x for x in eles]
                element_terms.extend(eles)
                if self.declose(term) == term:
                    if term == '...':
                        n_ele = data.n_space_dim - len([t for t in shape_terms if self.declose(t) == t]) + 1
                    else: n_ele = 1
                    ref_shape += bt.Size(-1) * n_ele
                else:
                    n_ele = len(eles)
                    if '...' in eles: n_ele = getattr(data, 'n_%s_dim'%{'{}':'batch','[]':'feature','""':'sequence',"''":'sequence','()':'func'}[term[0] + term[-1]])
                    ref_shape += bt.Size(eval(term[0] + ', '.join(['-1'] * n_ele) + term[-1]))
            data_shape = data.shape
            if data_shape.has_feature: ref_shape = ref_shape.with_feature(data.feature)
            elif opt_shape.has_feature: ref_shape = ref_shape.with_feature(tuple())
            if data_shape.has_sequence: ref_shape = ref_shape.with_sequence(data.sequence)
            elif opt_shape.has_sequence: ref_shape = ref_shape.with_sequence(tuple())
            if data_shape.has_space: ref_shape = ref_shape.with_space(data.space)
            elif opt_shape.has_space: ref_shape = ref_shape.with_space(tuple())
            data_shape = data.unsqueeze_to(ref_shape).shape
            error = TypeError(f"'{data_name}' with shape {data_shape} does not follow shape {shape_expr}")
            if '...' in element_terms:
                i = element_terms.index('...')
                avouch(0 < i < len(element_terms) - 1, TypeError(f"Ellipsis should be between two elements, not ({shape_expr})"))
                prev = element_terms[i - 1]
                post = element_terms[i + 1]
                _, prev_index = prev.split('_', 1)
                _, post_index = post.split('_', 1)
                prev_index = self.declose(prev_index)
                post_index = self.declose(post_index)
                variable_names.extend(self.get_variables(prev_index))
                variable_names.extend(self.get_variables(post_index))
                
                self.record_expr_value(element_terms[:i], data_shape)
                n_omit = len(data_shape) - len(element_terms) + 3
                avouch(n_omit > 1, error)
                self.record_expr_value(post_index + '-' + prev_index + '+1', n_omit)
                self.record_expr_value(element_terms[i + 1:], data_shape[i + n_omit - 2:])
                n_omits[data_name] = n_omit
            else:
                avouch(len(element_terms) == len(data_shape), error)
                self.record_expr_value(element_terms, data_shape)
                
        if len(variable_names) == 0:
            if any(eval(expr) != value for expr, value in self.exp_val_pair): raise error
            
        else:
        
            # solve the collected equations
            variable_names = list(set(variable_names))
            variables = symbols(' '.join(variable_names))
            exec(', '.join(variable_names) + ' = variables')
            equations = []
            for expr, value in self.exp_val_pair:
                equations.append(Eq(eval(expr), value))
            for ceq in const_eqs:
                equations.append(Eq(eval(ceq), 0))
            variable_values = solve(equations, variables, dict=True)
            exec('del ' + ', '.join(variable_names))
            variable_values = [{k.name: int(v) for k, v in vs.items()} for vs in variable_values if all(x.is_Integer and x > 0 for x in vs.values())]
            
            if len(variable_values) > 1:
                names = [n for n in variable_values[0] if len(set([vs[n] for vs in variable_values])) > 1]
                raise TypeError(f"{names[0] if len(names) == 1 else names} has multiple possible values {list(set([vs[n] for vs in variable_values]))} for {shape_expr} = {data_shape}")
            elif len(variable_values) == 0: raise error
            variable_values = variable_values[0]
            
            # try to solve the optional equations (due to 1 values)
            variable_names = [n for n in variable_names if n not in variable_values]
            if len(variable_names) > 0:
                new_exp_val_pair = []
                for expr, value in self.opt_exp_val_pair:
                    for k, v in variable_values.items():
                        while True:
                            res = re.search(rf'([^A-Za-z0-9]|^)({k})([^A-Za-z0-9]|$)', expr)
                            if res is None: break
                            i, j = res.span(2)
                            expr = expr[:i] + str(v) + expr[j:]
                    if self.get_variables(expr):
                        new_exp_val_pair.append((expr, value))
                locals().update(variable_values)
                variables = symbols(' '.join(variable_names))
                exec(', '.join(variable_names) + ' = variables')
                equations = []
                for expr, value in new_exp_val_pair:
                    equations.append(Eq(eval(expr), value))
                new_variable_values = solve(equations, variables, dict=True)
                exec('del ' + ', '.join(variable_names))
                if len(new_variable_values) == 1: variable_values.update({k.name: int(v) for k, v in new_variable_values[0].items() if v.is_Integer and v > 0})
            
            for k, v in self.default_values:
                if k not in variable_values: variable_values[k] = v
            for k, v in variable_values.items(): self.vars.globals[k] = v
            locals().update(variable_values)
        
        for data_name, shape_expr in kwargs.items():
            shape_expr, *_ = self.split(shape_expr, sep='')
            shape_expr = shape_expr.replace(':optional', '')
            data = self.vars[data_name]
            if data is None: continue
            if '...' in shape_expr:
                terms = [x.strip() for x in shape_expr.strip('()').split(',')]
                i = terms.index('...')
                if i > 1: left_size = bt.Size(*eval(','.join(terms[:i-1])))
                else: left_size = bt.Size()
                if i < len(terms) - 2: right_size = bt.Size(*eval(','.join(terms[i+2:])))
                else: right_size = bt.Size()
                size = left_size + (-1,) * n_omits[data_name] + right_size
            else: size = bt.Size(*eval(shape_expr))
            self.vars[data_name] = data.expand_to(size)
            
    def usage(self, index=None, /, **kwargs):
        self.index += 1
        if index is None: index = self.index
        try:
            self.set(**kwargs)
            self.vars.globals['usage'] = index
        except TypeError: ...
        return self
    
def as_shape(self, shape_expr: str):
    env_vars = get_environ_vars()
    loc_vars = env_vars.locals
    shape_list = [x.strip() for x in shape_expr.strip("()").split(',')]
    ibatch, ichannel = None, None
    old_shape = self.shape
    dim_pos = [0] * len(shape_list)
    pointer = 0
    for i, x in enumerate(shape_list):
        if x == '...': break
        if x.startswith('[') and x.endswith(']'): dim_pos[i] = old_shape.batch_dimension; x = x.strip('[]')
        if x.startswith('{') and x.endswith('}'): dim_pos[i] = old_shape.channel_dimension; x = x.strip('{}')
        if x.endswith(":optional"): x = x[:-len(":optional")]
        x = get_snakename(x)
        if x in loc_vars: x = loc_vars[x]

@alias("upper_triangular_matrix")
@batorch_wrapper
def upper_matrix(elements: bt.Tensor):
    '''
    elements: ({n_batch}, [n_dim * (n_dim + 1) / 2])
    output: ({n_batch}, [n_dim, n_dim])
    Compute the matrix with upper triangular elements of `elements`:
        For nxn matrix space (n_dim = n), the input is 𝜔:({n_batch}, [m = n(n+1)/2])
            The output is in shape ({n_batch}, [n, n]) of matrices:
                [[𝜔ₘ 𝜔ₘ₋₁ ... 𝜔ₘ₋ₙ₊₂  𝜔ₘ₋ₙ₊₁ ]
                 [0  𝜔ₘ₋ₙ ... 𝜔ₘ₋₂ₙ₊₄ 𝜔ₘ₋₂ₙ₊₃]
                 [︙  ︙   ⋱    ︙       ︙   ]
                 [0   0   ...   𝜔₃      𝜔₂   ]
                 [0   0   ...   0       𝜔₁   ]]
        ...
    '''
    avouch(elements.n_feature_dim == 1, TypeError("elements for 'upper_matrix' should be a tensor of shape ({n_batch}, [n_dim * (n_dim + 1) / 2])"))
    if not elements.has_batch: elements = elements.unsqueeze({})
    n_batch = elements.n_batch
    n_dim = int(math.sqrt(2 * elements.n_channel))
    elements = elements.flip([])
    output = bt.zeros({n_batch}, [n_dim, n_dim], dtype=elements.dtype, device=elements.device)
    row_len = bt.arange(n_dim, 0, -1)
    for i in range(n_dim): output[:, i, i:] = elements[:, row_len[:i].sum():row_len[:(i+1)].sum()]
    return output

@alias("lower_triangular_matrix")
@batorch_wrapper
def lower_matrix(elements: bt.Tensor):
    '''
    elements: ({n_batch}, [n_dim * (n_dim + 1) / 2])
    output: ({n_batch}, [n_dim, n_dim])
    Compute the matrix with lower triangular elements of `elements`:
        For nxn matrix space (n_dim = n), the input is 𝜔:({n_batch}, [m = n(n+1)/2])
            The output is in shape ({n_batch}, [n, n]) of matrices:
                [[ 𝜔ₘ       0     ...    0   0 ]
                 [𝜔ₘ₋₁    𝜔ₘ₋ₙ    ...    0   0 ]
                 [ ︙       ︙      ⋱   ︙   ︙ ]
                 [𝜔ₘ₋ₙ₊₂  𝜔ₘ₋₂ₙ₊₄  ...   𝜔₃  0  ]
                 [𝜔ₘ₋ₙ₊₁  𝜔ₘ₋₂ₙ₊₃  ...   𝜔₂  𝜔₁ ]]
        ...
    '''
    avouch(elements.n_feature_dim == 1, TypeError("elements for 'upper_matrix' should be a tensor of shape ({n_batch}, [n_dim * (n_dim + 1) / 2])"))
    if not elements.has_batch: elements = elements.unsqueeze({})
    n_batch = elements.n_batch
    n_dim = int(math.sqrt(2 * elements.n_channel))
    elements = elements.flip([])
    output = bt.zeros({n_batch}, [n_dim, n_dim], dtype=elements.dtype, device=elements.device)
    row_len = bt.arange(n_dim, 0, -1)
    for i in range(n_dim): output[:, i, i:] = elements[:, row_len[:i].sum():row_len[:(i+1)].sum()]
    return output.T

@alias("skew_symmetric")
@batorch_wrapper
def cross_matrix(axis: bt.Tensor):
    '''
    axis: ({n_batch}, [n_dim * (n_dim - 1) / 2])
    output: ({n_batch}, [n_dim, n_dim])
    Compute the skew symmetric matrix:
        For 2x2 matrix space (n_dim = 2), the input is 𝜗:({n_batch}, [1])
            The output is in shape ({n_batch}, [2, 2]) of matrices:
                [[0  𝜗]
                 [-𝜗 0]]
        For 3x3 matrix space (n_dim = 3), the input is 𝜔:({n_batch}, [3])
            The output is in shape ({n_batch}, [3, 3]) of matrices:
                [[ 0   𝜔₃ -𝜔₂]
                 [-𝜔₃  0   𝜔₁]
                 [ 𝜔₂ -𝜔₁  0 ]]
        For 4x4 matrix space (n_dim = 4), the input is 𝜔:({n_batch}, [6])
            The output is in shape ({n_batch}, [4, 4]) of matrices:
                [[ 0   𝜔₆ -𝜔₅  𝜔₄]
                 [-𝜔₆  0   𝜔₃ -𝜔₂]
                 [ 𝜔₅ -𝜔₃  0   𝜔₁]
                 [-𝜔₄  𝜔₂ -𝜔₁  0 ]]
        ...
    '''
    avouch(axis.n_feature_dim == 1, TypeError("axis for 'cross_matrix' should be a tensor of shape ({n_batch}, [n_dim * (n_dim - 1) / 2])"))
    if not axis.has_batch: axis = axis.unsqueeze({})
    n_batch = axis.n_batch
    n_dim = int(math.sqrt(2 * axis.n_channel)) + 1
    axis = axis.flip([])
    upper = bt.zeros({n_batch}, [n_dim, n_dim], dtype=axis.dtype, device=axis.device)
    row_len = bt.arange(n_dim - 1, 0, -1)
    row_sign = bt.where(row_len % 2 == 0, bt.ones_like(row_len), -bt.ones_like(row_len)).as_feature_tensor()
    for i in range(n_dim):
        upper[:, i, i+1:] = axis[:, row_len[:i].sum():row_len[:i+1].sum()] * row_sign[:(n_dim - i - 1)]
    return upper - upper.T

@alias("skew_symmetric_params")
@batorch_wrapper
def uncross_matrix(cross_matrix: bt.Tensor):
    '''
    cross_matrix: ({n_batch}, [n_dim, n_dim])
    output: ({n_batch}, [n_dim * (n_dim - 1) / 2])
    The inverse operation of cross_matrix. 
    '''
    avouch(cross_matrix.n_feature_dim == 2 or not cross_matrix.has_feature and cross_matrix.n_space_dim == 2, 
           TypeError("cross_matrix for 'cross_matrix' should be a tensor of shape ({n_batch}, [n_dim, n_dim])"))
    if not cross_matrix.has_batch: cross_matrix = cross_matrix.unsqueeze({})
    if not cross_matrix.has_feature:
        cross_matrix.sz_feature_dim_(cross_matrix.n_space_dim if cross_matrix.sz_sequence_dim < 0 else -cross_matrix.n_space_dim)
    avouch(cross_matrix.feature[0] == cross_matrix.feature[1], 
           TypeError("cross_matrix for 'cross_matrix' should be a tensor of shape ({n_batch}, [n_dim, n_dim])"))
    n_batch = cross_matrix.n_batch
    n_dim = cross_matrix.feature[0]
    axis = bt.zeros({n_batch}, [n_dim * (n_dim - 1) // 2], dtype=cross_matrix.dtype, device=cross_matrix.device)
    row_len = bt.arange(n_dim - 1, 0, -1)
    row_sign = bt.where(row_len % 2 == 0, bt.ones_like(row_len), -bt.ones_like(row_len)).as_feature_tensor()
    for i in range(n_dim):
        axis[:, row_len[:i].sum():row_len[:i+1].sum()] = cross_matrix[:, i, i+1:] * row_sign[:(n_dim - i - 1)]
    axis = axis.flip([])
    return axis

class summary:
    def __init__(self, *Xs, show_thumb = False, thumb_size = None, plot = False):
        self.smys = []
        self.show_as_plot = plot
        if plot: self.canvases = []
        if not hasattr(self, 'var_names'):
            args_exp = get_args_expression("summary")
            tokens = tokenize(args_exp, sep=', ')
            self.var_names = [vn.strip() for vn in tokens if len(tokenize(vn, sep='=')) == 1]
        avouch(len(self.var_names) <= len(Xs), f"Too many variable names recognized: {self.var_names}")
        starts_with_star = [i for i, x in enumerate(self.var_names) if x.startswith('*')]
        n_starts_with_star = 0
        if len(starts_with_star) > 1:
            avouch(None, f"Too much expandable variable in {self.var_names}: more than one. ")
        elif len(starts_with_star) == 0:
            avouch(len(self.var_names) == len(Xs), f"No expandable variable in {self.var_names} but detected more input arguments. ")
        else:
            starts_with_star = starts_with_star[0]
            n_starts_with_star = len(Xs) - len(self.var_names) + 1
        
        if thumb_size is None: thumb_size = 4
        else: show_thumb = True
        self.show_thumb = show_thumb
        self.thumb_size = thumb_size
        
        for i, x in enumerate(Xs):
            if n_starts_with_star > 0:
                if 0 <= i - starts_with_star < n_starts_with_star: j = i - starts_with_star; i = starts_with_star
                elif i >= starts_with_star + n_starts_with_star: i = i - n_starts_with_star
            var_name = self.var_names[i] + (f'[{j}]' if i == starts_with_star else '')
            class_name = x.__class__.__name__
            g = None
            if not isinstance(x, bt.torch.Tensor):
                l = touch(lambda: len(x), default=0)
                if l <= 10:
                    self.smys.append(f">>{var_name}<< " + str(x) + '\n')
                    continue
                else: x = bt.tensor(x)
            elif not isinstance(x, bt.Tensor):
                if x.is_leaf and x.grad is not None: g = bt.tensor(x.grad.clone().detach())
                x = bt.tensor(x)
            special_dim_str = f"[{x.n_special_dim}]+" if x.n_special_dim > 0 else ''
            self.out = SPrint(f">>Summary of {var_name}<<\n")
            def size_str(Size=None, Dimension=None, Class=None):
                return f"{f'[{Size.n_special_dim}]+' if Size.n_special_dim > 0 else ''}{Size.n_space_dim}D {Class.__name__} of size {self.__raw_str__(Size)}"
            self.__add_prop__('Size', value2str_func = size_str, Size=x.shape, Dimension=x.n_dim, Class=x.__class__)
            self.__add_prop__('Device', x.device, lambda d: 'cpu' if d.type == 'cpu' else f"{d.type}:{d.index}")
            self.__add_prop__('Dtype', x.dtype, self.__raw_str__)
            self.__add_prop__('Requires_Gradient', x.requires_grad)
            if x.requires_grad: self.__add_prop__('Gradient_Func', x.grad_fn)

            self.show_grad = True
            self.__display_values__(x)
            if self.show_grad and g is not None:
                self.__add_prop__('Gradient_Size', g.shape, self.__raw_str__)
                self.__display_values__(x, is_grad=True)
            self.smys.append(self.out.text)
            if plot:
                while x.n_dim > 2 and not (x.n_dim == 3 and x.size(-1) == 3):
                    if x.has_special:
                        d_sample = x.special_dims[0]
                    else: d_sample = argmin(list(x.shape))[0]
                    x = x.pick(x.size(d_sample) // 2, d_sample)
                self.canvases.append(x.detach())
                
    def __add_prop__(self, name, value=null, value2str_func=None, prompt=None, display_only=False, **assign_values):
        if prompt is not None: prompt = f" ({prompt})"
        else: prompt = ''
        if value is not null:
            if value2str_func is not None: value_str = value2str_func(value)
            else: value_str = str(value)
            if not display_only: setattr(self, name, value)
        else:
            avouch(value2str_func is not None, TypeError("'value2str_func' should be specified when multiple values are given for 'summary.__add_prop__'. "))
            value_str = value2str_func(**assign_values)
            if not display_only:
                for n, v in assign_values.items(): setattr(self, n, v)
        tag = f"【{name}{prompt}】"
        n_align = math.ceil((len(tag) + 2) / 25)
        self.out(tag + ' ' * (n_align * 25 - len(tag) - 2) + value_str)
             
    def __raw_str__(self, x):
        sx = str(x)
        if 'Size(' in sx: return sx.split('Size')[-1]
        return tokenize(tokenize(sx.split('(', 1)[-1], sep=[',', ')'])[0], sep='.')[-1]

    def __approx_quantile__(self, x):
        valid_dims = bt.tensor(x.shape).cumprod(0) >> 24 < 1
        if all(valid_dims): x = x.flatten()
        elif all(~valid_dims):
            n_replicate = (x.n_ele >> 24) + 1
            n_sample = x.n_ele // n_replicate
            x = x[:n_replicate * n_sample].reshape(n_sample, n_replicate)
        else: x = x.flatten(0, valid_dims.argmin().item()-1)
        x = x.float().quantile(bt.arange(6).float() / 5, dim=0)
        if x.n_dim > 1: x = x.mean(*range(1, x.n_dim))
        return x
    
    def __display_values__(self, x, is_grad=False):
        nan_mask = bt.isnan(x)
        num_nans = nan_mask.sum()
        inf_mask = bt.isinf(x)
        num_infs = inf_mask.sum()
        nan_inf_mask = nan_mask | inf_mask
        valid_nums = x[~nan_inf_mask]
        if num_nans + num_infs == 0: show_error = False; show_valid = True
        elif (num_nans + num_infs) / x.n_ele < 0.5: show_error = True; show_valid = True
        else: show_error = True; show_valid = False; self.show_grad = False
        if show_valid:
            values_str = lambda vs: f"(min:{vs[0]}, med:{vs[1]}, max:{vs[2]})"
            if x.is_complex():
                vre_name, vim_name, name = 'Values_Re', 'Values_Im', 'Values_Re/Im'
                if is_grad: vre_name, vim_name, name = ['Gradient_' + x for x in (vre_name, vim_name, name)]
                vre_val = (valid_nums.real.min(), valid_nums.real.median(), valid_nums.real.max())
                vim_val = (valid_nums.imag.min(), valid_nums.imag.median(), valid_nums.imag.max())
                self.__add_prop__(name, value2str_func=lambda **k: f"Re{values_str(k[vre_name])}, Im{values_str(k[vim_name])}", **{vre_name: vre_val, vim_name: vim_val})
            else:
                name = 'Gradient_Values' if is_grad else 'Values'
                val = (valid_nums.min(), valid_nums.median(), valid_nums.max())
                self.__add_prop__(name, val, values_str)
        if show_error:
            if num_nans > 0:
                nan_name = f"{num_nans} NaN"
                if num_nans > 1: nan_name += 's'
                has_nan_maps = nan_mask.sum(...) > 0
                num_nan_maps = has_nan_maps.sum()
                nan_mask = nan_mask[has_nan_maps, ...][0]
                nan_center = bt.mask_center(nan_mask)
                if is_grad:
                    self.Gradient_NaN = self.Gradient_NaNs = num_nans
                else:
                    self.NaN = self.NaNs = num_nans
                    self.NaN_maps = num_nan_maps
                    self.NaN_center = tuple(nan_center)
                self.__add_prop__(nan_name, value2str_func=lambda **_: f"{num_nan_maps}/{has_nan_maps.n_ele} maps with nan, e.g. center: {self.__raw_str__(nan_center)}", display_only=True)
            if num_infs > 0:
                inf_name = f"{num_infs} Inf"
                if num_infs > 1: inf_name += 's'
                has_inf_maps = inf_mask.sum(...) > 0
                num_inf_maps = has_inf_maps.sum()
                inf_mask = inf_mask[has_inf_maps, ...][0]
                inf_center = bt.mask_center(inf_mask)
                if is_grad:
                    self.Gradient_Inf = self.Gradient_Infs = num_infs
                else:
                    self.Inf = self.Infs = num_infs
                    self.Inf_maps = num_inf_maps
                    self.Inf_center = tuple(inf_center)
                self.__add_prop__(inf_name, value2str_func=lambda **_: f"{num_inf_maps}/{has_inf_maps.n_ele} maps with inf, e.g. center: {self.__raw_str__(inf_center)}", display_only=True)
        if show_valid:
            approx_q = x.n_ele >> 24 > 0
            self.__add_prop__('Gradient_Quantiles' if is_grad else 'Quantiles', self.__approx_quantile__(valid_nums), 
                              lambda q: ('\n >> ' if approx_q else '') + f"([0%]{q[0]}, [20%]{q[1]}, [40%]{q[2]}, [60%]{q[3]}, [80%]{q[4]}, [100%]{q[5]})", 
                              prompt='approx. due to too many elements' if approx_q else None)
            values = valid_nums.unique()
            def range_str(values):
                if values.n_ele <= 10:
                    if values.n_ele > 6: return f"\n >> [{', '.join(str(x) for x in values)}]"
                    else:
                        num_pairs = [(v, (x == v).sum().item()) for v in values]
                        return f"\n >> [{', '.join(f'{v}({n})' for v, n in num_pairs)}]"
                else: return f"\n >> [{', '.join(str(x) for x in values[:5])}, ..., {', '.join(str(x) for x in values[-5:])}]"
            self.__add_prop__('Gradient_Range' if is_grad else 'Range', values, value2str_func=range_str, 
                              prompt=f"{values.n_ele} unique element" + ('s' if values.n_ele > 1 else ''))
        if self.show_thumb:
            gaps = bt.floor((bt.channel_tensor(list(x.space)) - 1) / (self.thumb_size - 1)).clamp(min=1).long()
            down_sample = x[(slice(None),) * x.space_start + (image_grid(*(self.thumb_size if s > self.thumb_size else max(s // g, 1) for s, g in zip(x.space, gaps))) * gaps).split(1, dim=[], squeezedim=True)]
            for i in down_sample.special_dims[::-1]: down_sample = down_sample.pick(0, dim=i)
            self.__add_prop__('Gradient_Thumb' if is_grad else 'Thumb', down_sample, 
                              value2str_func=lambda x: '\n >> ' + bt.Tensor.__shift_repr__(x.__raw_repr__(), 4), 
                              prompt=f"downsampled to {self.thumb_size}{f' x {self.thumb_size}'*(x.n_space_dim-1)}")
    
    def show(self):
        if not self.show_as_plot:
            print(self)
            return
        plt.figure(figsize=(15, 7))
        
        n_plots = len(self.canvases)
        self.subplot_rows = math.floor(math.sqrt(n_plots))
        self.subplot_cols = math.ceil(n_plots / self.subplot_rows)

        old_minus = plt.rcParams['axes.unicode_minus']
        old_font = plt.rcParams['font.sans-serif']
        font = 'Heiti TC'
        if font not in plt.rcParams['font.sans-serif']:
            plt.rcParams['font.sans-serif'].insert(0, font)
        plt.rcParams['axes.unicode_minus'] = False
        for i, x in enumerate(self.canvases):
            ax = plt.subplot(self.subplot_rows, self.subplot_cols, i+1)
            old_size = plt.rcParams['font.size']
            plt.rcParams['font.size'] = 6
            if x.n_dim >= 2:
                img = plt.imshow(x, cmap=plt.cm.gray)
                plt.colorbar(mappable=img, ax=ax, orientation='vertical')
                plt.text(-0.5, -0.5, self.smys[i])
            elif x.n_dim == 1:
                plt.plot(bt.arange(len(x)), x)
                plt.text(0, plt.ylim()[1], self.smys[i])
            else: raise TypeError("Failed to display 0-dimensional data in plot.")
            plt.rcParams['font.size'] = old_size
        plt.rcParams['font.sans-serif'] = old_font
        plt.rcParams['axes.unicode_minus'] = old_minus 
        plt.show()

    @alias("__repr__")
    def __str__(self): return ''.join(self.smys)

class display(summary):
    def __init__(self, *args, **kwargs):
        kwargs['plot'] = True
        args_exp = get_args_expression("display")
        tokens = tokenize(args_exp, sep=', ')
        self.var_names = [vn.strip() for vn in tokens if len(tokenize(vn, sep='=')) == 1]
        return super().__init__(*args, **kwargs)
    
def num_model_params(model):
    s = ByteSize(0)
    for param in model.parameters():
        s += param.numel() * dtype(param.dtype).bits // 8
    return s
